from .base_x_lumped import BaseModel
from .x_lumped_no_current_collectors import NoCurrentCollector
from .x_lumped_0D_current_collectors import CurrentCollector0D
from .x_lumped_1D_current_collectors import CurrentCollector1D
from .x_lumped_2D_current_collectors import CurrentCollector2D
from .x_lumped_1D_set_temperature import SetTemperature1D



#
# Print parameters
#
import pybamm
from collections import defaultdict

# Set list of attributes to ignore, for when we are evaluating parameters from a class
# of parameters
ignore = [
    "__name__",
    "__doc__",
    "__package__",
    "__loader__",
    "__spec__",
    "__file__",
    "__cached__",
    "__builtins__",
    "absolute_import",
    "division",
    "print_function",
    "unicode_literals",
    "pybamm",
    "constants",
    "np",
]


def print_parameters(parameters, parameter_values, output_file=None):
    """
    Return dictionary of evaluated parameters, and optionally print these evaluated
    parameters to an output file.
    For dimensionless parameters that depend on the C-rate, the value is given as a
    function of the C-rate (either x * Crate or x / Crate depending on the dependence)

    Parameters
    ----------
    parameters : class or dict containing :class:`pybamm.Parameter` objects
        Class or dictionary containing all the parameters to be evaluated
    parameter_values : :class:`pybamm.ParameterValues`
        The class of parameter values
    output_file : string, optional
        The file to print parameters to. If None, the parameters are not printed, and
        this function simply acts as a test that all the parameters can be evaluated,
        and returns the dictionary of evaluated parameters.

    Returns
    -------
    evaluated_parameters : defaultdict
        The evaluated parameters, for further processing if needed

    Notes
    -----
    A C-rate of 1 C is the current required to fully discharge the battery in 1 hour,
    2 C is current to discharge the battery in 0.5 hours, etc
    """
    # If 'parameters' is a class, extract the dict
    if not isinstance(parameters, dict):
        parameters = {k: v for k, v in parameters.__dict__.items() if k not in ignore}

    evaluated_parameters = defaultdict(list)
    # Calculate parameters for each C-rate
    for Crate in [1, 10]:
        # Update Crate
        parameter_values.update({"C-rate": Crate}, check_already_exists=False)
        for name, symbol in parameters.items():
            if not callable(symbol):
                proc_symbol = parameter_values.process_symbol(symbol)
                if not (
                    callable(proc_symbol)
                    or any(
                        isinstance(x, (pybamm.Concatenation, pybamm.Broadcast))
                        for x in proc_symbol.pre_order()
                    )
                ):
                    evaluated_parameters[name].append(proc_symbol.evaluate(t=0))

    # Calculate C-dependence of the parameters based on the difference between the
    # value at 1C and the value at C / 10
    for name, values in evaluated_parameters.items():
        if values[1] == 0 or abs(values[0] / values[1] - 1) < 1e-10:
            C_dependence = ""
        elif abs(values[0] / values[1] - 10) < 1e-10:
            C_dependence = " * Crate"
        elif abs(values[0] / values[1] - 0.1) < 1e-10:
            C_dependence = " / Crate"
        evaluated_parameters[name] = (values[0], C_dependence)
    # Print the evaluated_parameters dict to output_file
    if output_file:
        print_evaluated_parameters(evaluated_parameters, output_file)

    return evaluated_parameters


def print_evaluated_parameters(evaluated_parameters, output_file):
    """
    Print a dictionary of evaluated parameters to an output file

    Parameters
    ----------
    evaluated_parameters : defaultdict
        The evaluated parameters, for further processing if needed
    output_file : string, optional
        The file to print parameters to. If None, the parameters are not printed, and
        this function simply acts as a test that all the parameters can be evaluated

    """
    # Get column width for pretty printing
    column_width = max(len(name) for name in evaluated_parameters.keys())
    s = "{{:>{}}}".format(column_width)
    with open(output_file, "w") as file:
        for name, (value, C_dependence) in sorted(evaluated_parameters.items()):
            if 0.001 < abs(value) < 1000:
                file.write((s + " : {:10.4g}{!s}\n").format(name, value, C_dependence))
            else:
                file.write((s + " : {:10.3E}{!s}\n").format(name, value, C_dependence))



#
# Inverse Bulter-Volmer class
#
import pybamm
from .base_inverse_kinetics import BaseInverseKinetics
from ..kinetics.butler_volmer import ButlerVolmer


class InverseButlerVolmer(BaseInverseKinetics, ButlerVolmer):
    """
    A base submodel that implements the inverted form of the Butler-Volmer relation to
    solve for the reaction overpotential.

    Parameters
    ----------
    param
        Model parameters
    domain : iter of str, optional
        The domain(s) in which to compute the interfacial current. Default is None,
        in which case j.domain is used.

    **Extends:** :class:`pybamm.interface.kinetics.ButlerVolmer`

    """

    def __init__(self, param, domain):
        super().__init__(param, domain)

    def _get_overpotential(self, j, j0, ne, T):
        return (2 * (1 + self.param.Theta * T) / ne) * pybamm.arcsinh(j / (2 * j0))



#
# Class for reaction driven porosity changes
#
import pybamm
from .base_porosity import BaseModel


class Full(BaseModel):
    """Full model for reaction-driven porosity changes

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel


    **Extends:** :class:`pybamm.porosity.BaseModel`
    """

    def __init__(self, param):
        super().__init__(param)

    def get_fundamental_variables(self):

        eps = pybamm.standard_variables.eps
        variables = self._get_standard_porosity_variables(eps)
        return variables

    def get_coupled_variables(self, variables):

        j_n = variables["Negative electrode interfacial current density"]
        j_p = variables["Positive electrode interfacial current density"]

        deps_dt_n = -self.param.beta_surf_n * j_n
        deps_dt_s = pybamm.FullBroadcast(
            0, "separator", auxiliary_domains={"secondary": "current collector"}
        )
        deps_dt_p = -self.param.beta_surf_p * j_p

        deps_dt = pybamm.Concatenation(deps_dt_n, deps_dt_s, deps_dt_p)

        variables.update(self._get_standard_porosity_change_variables(deps_dt))

        return variables

    def set_rhs(self, variables):

        eps = variables["Porosity"]
        deps_dt = variables["Porosity change"]

        self.rhs = {eps: deps_dt}

    def set_initial_conditions(self, variables):
        eps = variables["Porosity"]
        self.initial_conditions = {eps: self.param.epsilon_init}



from pybamm import exp


def graphite_electrolyte_reaction_rate_Dualfoil1998(T, T_inf, E_r, R_g):
    """
    Reaction rate for Butler-Volmer reactions between graphite and LiPF6 in EC:DMC.

    References
    ----------
    .. [2] http://www.cchem.berkeley.edu/jsngrp/fortran.html

    Parameters
    ----------
    T: :class: `numpy.Array`
        Dimensional temperature
    T_inf: double
        Reference temperature
    E_r: double
        Reaction activation energy
    R_g: double
        The ideal gas constant

    Returns
    -------
    :`numpy.Array`
        Reaction rate
    """
    m_ref = 2 * 10 ** (-5)
    arrhenius = exp(E_r / R_g * (1 / T_inf - 1 / T))

    return m_ref * arrhenius



import os
import pybamm
import argparse
import shutil

from pathlib import Path


def yes_or_no(question):
    while "Please answer yes(y) or no(n).":
        reply = str(input(question + " (y/n): ")).lower().strip()
        if reply[:1] == "y":
            return True
        if reply[:1] == "n":
            return False


def get_parser(description):
    """
    Create a parser with the following arguments:
    parameter_dir: str
    battery_type: str
    component: str
    force: bool, optional

    Parameter:
    ----------
    description: A description of the command [str]

    Returns
    -------
    argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "parameter_dir", type=str, help="Name of the parameter directory",
    )
    parser.add_argument("battery_type", choices=["lithium-ion", "lead-acid"])
    parser.add_argument(
        "component",
        choices=[
            "anodes",
            "cathodes",
            "cells",
            "electrolytes",
            "experiments",
            "separators",
        ],
    )
    parser.add_argument("-f", "--force", action="store_true")
    return parser


def copy_directory(source, destination, overwrite):
    """
    Copy directory structure source as destination, dealing
    with overwriting.

    Parameters:
    -----------
    source: Directory structure [str]
    destination: Directory structure [str]
    overwrite: Whether or not to overwrite [bool]

    Example:
    --------
    source = "site-packages/pybamm/input/parameters/lead-acid/anodes/blablah"
    destination = "input/parameters/lead-acid/anodes/blablah"
    Will copy
      "input/parameters/lead-acid/anodes/blablah"
    in current working directory.
    """
    try:
        shutil.copytree(source, destination)
    except FileExistsError:
        if overwrite:
            shutil.rmtree(destination)
            shutil.copytree(source, destination)
        elif yes_or_no("Parameter set already defined, erase?"):
            shutil.rmtree(destination)
            shutil.copytree(source, destination)
        else:
            print("Doing nothing.")
    # Implementation note:
    # In python 3.7 shutil.copytree() does not provide an option to
    # automatically overwrite a directory tree.
    # When switching to python 3.8, the following function could be
    # written using
    # shutil.copytree(args.parameter_dir, destination_dir, exist_ok=True)
    # and getting rid of the rmtree calls.
    # see https://docs.python.org/3/library/shutil.html


def add_parameter(arguments=None):
    """
    Add a parameter directory to package input directory.
    This allows the parameters to be used from anywhere in the system.

    Example:
    "add_parameter foo lithium-ion anodes" will copy directory foo in
    "pybamm/input/parameters/lithium-ion/anodes".
    """
    parser = get_parser("Copy parameter to the PyBaMM package directory.")
    args = parser.parse_args(arguments)

    parameters_root_dir = os.path.join(pybamm.__path__[0], "input", "parameters")

    parameter_dir_name = Path(args.parameter_dir).name
    destination_dir = os.path.join(
        parameters_root_dir, args.battery_type, args.component, parameter_dir_name
    )

    copy_directory(args.parameter_dir, destination_dir, args.force)
    print("Copied {} to {}".format(args.parameter_dir, destination_dir))


def edit_parameter(arguments=None):
    """
    Copy a given parameter package directory to the current working directory
    for editing. The copy preserves the directory structure within the "input"
    directory, i.e

    ``edit_param(["graphite_Kim2011","lithium-ion","anodes"])``

    will create the directory structure
    "input/parameters/lithium-ion/anodes/graphite_Kim2011"
    in the current working directory.
    """
    parser = get_parser(
        "Pull parameter directory dir to current working directory for editing."
    )
    args = parser.parse_args(arguments)

    path = os.path.join(
        "input", "parameters", args.battery_type, args.component, args.parameter_dir
    )

    source_dir = os.path.join(pybamm.__path__[0], path)
    copy_directory(source_dir, path, args.force)


def list_parameters(arguments=None):
    """
    Output a list of available parameter sets for a given chemistry and component.
    The list is divided into package parameter serts and local parameter sets,
    located in the current working directory.

    >>> from pybamm.parameters_cli import list_parameters
    >>> list_parameters(["lithium-ion", "anodes"])
    Available package parameters:
      * graphite_Chen2020
      * graphite_mcmb2528_Marquis2019
      * graphite_Kim2011
    Available local parameters:
    """
    parser = argparse.ArgumentParser(
        description="List available parameter sets for a given chemistry and component."
    )
    parser.add_argument("battery_type", choices=["lithium-ion", "lead-acid"])
    parser.add_argument(
        "component",
        choices=[
            "anodes",
            "cathodes",
            "cells",
            "electrolytes",
            "experiments",
            "separators",
        ],
    )

    args = parser.parse_args(arguments)

    package_dir = os.path.join(
        pybamm.__path__[0], "input", "parameters", args.battery_type, args.component
    )
    root, package_dirs, files = next(os.walk(package_dir))

    print("Available package parameters:")
    for dirname in package_dirs:
        print("  * {}".format(dirname))

    local_dir = os.path.join("input", "parameters", args.battery_type, args.component)
    if os.path.isdir(local_dir):
        root, local_dirs, files = next(os.walk(local_dir))
    else:
        local_dirs = []
    print("Available local parameters:")
    for dirname in local_dirs:
        print("  * {}".format(dirname))



from .base_stefan_maxwell_conductivity import BaseModel
from .base_higher_order_stefan_maxwell_conductivity import BaseHigherOrder
from .leading_stefan_maxwell_conductivity import LeadingOrder
from .first_order_stefan_maxwell_conductivity import FirstOrder
from .composite_stefan_maxwell_conductivity import Composite
from .full_stefan_maxwell_conductivity import Full

from . import surface_potential_form



#
# Lithium-ion interface classes
#
from .base_interface import BaseInterface
from . import inverse_kinetics, kinetics
import pybamm


class BaseInterfaceLithiumIon(BaseInterface):
    """
    Base lthium-ion interface class

    Parameters
    ----------
    param :
        model parameters
    domain : str
        The domain to implement the model, either: 'Negative' or 'Positive'.


    **Extends:** :class:`pybamm.interface.BaseInterface`
    """

    def __init__(self, param, domain):
        super().__init__(param, domain)
        self.reaction_name = ""  # empty reaction name, assumed to be the main reaction

    def _get_exchange_current_density(self, variables):
        """
        A private function to obtain the exchange current density for a lithium-ion
        deposition reaction.

        Parameters
        ----------
        variables: dict
        `   The variables in the full model.

        Returns
        -------
        j0 : :class: `pybamm.Symbol`
            The exchange current density.
        """
        c_s_surf = variables[self.domain + " particle surface concentration"]
        c_e = variables[self.domain + " electrolyte concentration"]
        T = variables[self.domain + " electrode temperature"]

        # If variable was broadcast, take only the orphan
        if (
            isinstance(c_s_surf, pybamm.Broadcast)
            and isinstance(c_e, pybamm.Broadcast)
            and isinstance(T, pybamm.Broadcast)
        ):
            c_s_surf = c_s_surf.orphans[0]
            c_e = c_e.orphans[0]
            T = T.orphans[0]

        if self.domain == "Negative":
            prefactor = self.param.m_n(T) / self.param.C_r_n

        elif self.domain == "Positive":
            prefactor = self.param.gamma_p * self.param.m_p(T) / self.param.C_r_p

        j0 = prefactor * (
            c_e ** (1 / 2) * c_s_surf ** (1 / 2) * (1 - c_s_surf) ** (1 / 2)
        )

        return j0

    def _get_open_circuit_potential(self, variables):
        """
        A private function to obtain the open circuit potential and entropic change

        Parameters
        ----------
        variables: dict
            The variables in the full model.

        Returns
        -------
        ocp : :class:`pybamm.Symbol`
            The open-circuit potential
        dUdT : :class:`pybamm.Symbol`
            The entropic change in open-circuit potential due to temperature

        """
        c_s_surf = variables[self.domain + " particle surface concentration"]
        T = variables[self.domain + " electrode temperature"]

        # If variable was broadcast, take only the orphan
        if isinstance(c_s_surf, pybamm.Broadcast) and isinstance(T, pybamm.Broadcast):
            c_s_surf = c_s_surf.orphans[0]
            T = T.orphans[0]

        if self.domain == "Negative":
            ocp = self.param.U_n(c_s_surf, T)
            dUdT = self.param.dUdT_n(c_s_surf)

        elif self.domain == "Positive":
            ocp = self.param.U_p(c_s_surf, T)
            dUdT = self.param.dUdT_p(c_s_surf)

        return ocp, dUdT

    def _get_number_of_electrons_in_reaction(self):
        if self.domain == "Negative":
            ne = self.param.ne_n
        elif self.domain == "Positive":
            ne = self.param.ne_p
        return ne


class ButlerVolmer(BaseInterfaceLithiumIon, kinetics.ButlerVolmer):
    """
    Extends :class:`BaseInterfaceLithiumIon` (for exchange-current density, etc) and
    :class:`kinetics.ButlerVolmer` (for kinetics)
    """

    def __init__(self, param, domain):
        super().__init__(param, domain)


class InverseButlerVolmer(
    BaseInterfaceLithiumIon, inverse_kinetics.InverseButlerVolmer
):
    """
    Extends :class:`BaseInterfaceLithiumIon` (for exchange-current density, etc) and
    :class:`inverse_kinetics.InverseButlerVolmer` (for kinetics)
    """

    def __init__(self, param, domain):
        super().__init__(param, domain)



from pybamm import exp


def graphite_electrolyte_reaction_rate_Kim2011(T, T_inf, E_r, R_g):
    """
    Reaction rate for Butler-Volmer reactions between graphite and LiPF6 in EC:DMC
    [1].

    References
    ----------
    .. [1] Kim, G. H., Smith, K., Lee, K. J., Santhanagopalan, S., & Pesaran, A.
    (2011). Multi-domain modeling of lithium-ion batteries encompassing
    multi-physics in varied length scales. Journal of The Electrochemical
    Society, 158(8), A955-A969.

    Parameters
    ----------
    T: :class: `numpy.Array`
        Dimensional temperature
    T_inf: double
        Reference temperature
    E_r: double
        Reaction activation energy
    R_g: double
        The ideal gas constant

    Returns
    -------
    :`numpy.Array`
        Reaction rate
    """

    i0_ref = 36  # reference exchange current density at 100% SOC
    sto = 0.36  # stochiometry at 100% SOC
    c_s_n_max = 2.87 * 10 ** 4  # max electrode concentration
    c_s_n_ref = sto * c_s_n_max  # reference electrode concentration
    c_e_ref = 1.2 * 10 ** 3  # reference electrolyte concentration
    alpha = 0.5  # charge transfer coefficient

    m_ref = (
        2
        * i0_ref
        / (c_e_ref ** alpha * (c_s_n_max - c_s_n_ref) ** alpha * c_s_n_ref ** alpha)
    )

    arrhenius = exp(E_r / R_g * (1 / T_inf - 1 / T))

    return m_ref * arrhenius



#
# Class for x-lumped thermal submodel with 1D current collector
#
import pybamm

from .base_x_lumped import BaseModel


class CurrentCollector1D(BaseModel):
    """Class for x-lumped thermal model with 1D current collectors"""

    def __init__(self, param):
        super().__init__(param)

    def set_rhs(self, variables):
        T_av = variables["X-averaged cell temperature"]
        Q_av = variables["X-averaged total heating"]

        cooling_coeff = self._surface_cooling_coefficient()

        self.rhs = {
            T_av: (pybamm.laplacian(T_av) + self.param.B * Q_av + cooling_coeff * T_av)
            / self.param.C_th
        }

    def set_boundary_conditions(self, variables):
        T_av = variables["X-averaged cell temperature"]
        T_av_left = pybamm.boundary_value(T_av, "negative tab")
        T_av_right = pybamm.boundary_value(T_av, "positive tab")

        # Three boundary conditions here to handle the cases of both tabs at
        # the same side (top or bottom), or one either side. For both tabs on the
        # same side, T_av_left and T_av_right are equal, and the boundary condition
        # "no tab" is used on the other side.
        self.boundary_conditions = {
            T_av: {
                "negative tab": (
                    self.param.h * T_av_left / self.param.delta,
                    "Neumann",
                ),
                "positive tab": (
                    -self.param.h * T_av_right / self.param.delta,
                    "Neumann",
                ),
                "no tab": (pybamm.Scalar(0), "Neumann"),
            }
        }

    def _current_collector_heating(self, variables):
        """Returns the heat source terms in the 1D current collector"""
        phi_s_cn = variables["Negative current collector potential"]
        phi_s_cp = variables["Positive current collector potential"]
        Q_s_cn = self.param.sigma_cn_prime * pybamm.inner(
            pybamm.grad(phi_s_cn), pybamm.grad(phi_s_cn)
        )
        Q_s_cp = self.param.sigma_cp_prime * pybamm.inner(
            pybamm.grad(phi_s_cp), pybamm.grad(phi_s_cp)
        )
        return Q_s_cn, Q_s_cp

    def _yz_average(self, var):
        """Computes the y-z average by integration over z (no y-direction)"""
        return pybamm.z_average(var)



#
# Parameter classes
#
import numpy as np
import pybamm


class InputParameter(pybamm.Symbol):
    """A node in the expression tree representing an input parameter

    This node's value can be set at the point of solving, allowing parameter estimation
    and control

    Parameters
    ----------
    name : str
        name of the node

    """

    def __init__(self, name):
        super().__init__(name)

    def new_copy(self):
        """ See :meth:`pybamm.Symbol.new_copy()`. """
        return InputParameter(self.name)

    def _evaluate_for_shape(self):
        """
        Returns the scalar 'NaN' to represent the shape of a parameter.
        See :meth:`pybamm.Symbol.evaluate_for_shape()`
        """
        return np.nan

    def _jac(self, variable):
        """ See :meth:`pybamm.Symbol._jac()`. """
        return pybamm.Scalar(0)

    def _base_evaluate(self, t=None, y=None, u=None):
        # u should be a dictionary
        # convert 'None' to empty dictionary for more informative error
        if u is None:
            u = {}
        if not isinstance(u, dict):
            # if the special input "shape test" is passed, just return 1
            if u == "shape test":
                return 1
            raise TypeError("inputs u should be a dictionary")
        try:
            return u[self.name]
        # raise more informative error if can't find name in dict
        except KeyError:
            raise KeyError("Input parameter '{}' not found".format(self.name))



#
# Class for xyz-lumped thermal submodel with 2D current collectors
#
import pybamm

from .base_xyz_lumped import BaseModel


class CurrentCollector2D(BaseModel):
    """Class for xyz-lumped thermal submodel with 2D current collectors

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel


    **Extends:** :class:`pybamm.thermal.BaseModel`
    """

    def __init__(self, param):
        super().__init__(param)

    def _current_collector_heating(self, variables):
        """Returns the heat source terms in the 2D current collector"""
        phi_s_cn = variables["Negative current collector potential"]
        phi_s_cp = variables["Positive current collector potential"]

        Q_s_cn = self.param.sigma_cn_prime * pybamm.grad_squared(phi_s_cn)
        Q_s_cp = self.param.sigma_cp_prime * pybamm.grad_squared(phi_s_cp)
        return Q_s_cn, Q_s_cp

    def _surface_cooling_coefficient(self):
        """Returns the surface cooling coefficient in 2+1D"""
        # Note: assumes pouch cell geometry
        return (
            -2 * self.param.h / (self.param.delta ** 2) / self.param.l
            - 2 * (self.param.l_y + self.param.l_z) * self.param.h / self.param.delta
        )

    def _yz_average(self, var):
        """Computes the y-z average by integration over y and z"""
        return pybamm.yz_average(var)



#
# Class for x-lumped thermal submodel with 0D current collector
#
from .base_x_lumped import BaseModel


class CurrentCollector0D(BaseModel):
    """Class for x-lumped thermal model with 0D current collectors"""

    def __init__(self, param):
        super().__init__(param)

    def set_rhs(self, variables):
        T_av = variables["X-averaged cell temperature"]
        Q_av = variables["X-averaged total heating"]

        cooling_coeff = self._surface_cooling_coefficient()

        self.rhs = {
            T_av: (self.param.B * Q_av + cooling_coeff * T_av) / self.param.C_th
        }

    def _current_collector_heating(self, variables):
        """Returns the heat source terms in the 0D current collector"""
        i_boundary_cc = variables["Current collector current density"]
        Q_s_cn = i_boundary_cc ** 2 / self.param.sigma_cn
        Q_s_cp = i_boundary_cc ** 2 / self.param.sigma_cp
        return Q_s_cn, Q_s_cp

    def _yz_average(self, var):
        """In 1D volume-averaged quantities are unchanged"""
        return var



#
# Base class for particles each with uniform concentration (i.e. infinitely fast
# diffusion in r)
#
from ..base_particle import BaseParticle


class BaseModel(BaseParticle):
    """Base class for molar conservation in particles with uniform concentration
    in r (i.e. infinitely fast diffusion within particles).

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain of the model either 'Negative' or 'Positive'


    **Extends:** :class:`pybamm.particle.BaseParticle`
    """

    def __init__(self, param, domain):
        super().__init__(param, domain)

    def _unpack(self, variables):
        raise NotImplementedError

    def set_rhs(self, variables):

        c, _, j = self._unpack(variables)

        if self.domain == "Negative":
            self.rhs = {c: -3 * j / self.param.a_n}

        elif self.domain == "Positive":
            self.rhs = {c: -3 * j / self.param.a_p / self.param.gamma_p}



from pybamm import exp


def lico2_diffusivity_Dualfoil1998(sto, T, T_inf, E_D_s, R_g):
    """
    LiCo2 diffusivity as a function of stochiometry, in this case the
    diffusivity is taken to be a constant. The value is taken from Dualfoil [1].

    References
    ----------
    .. [1] http://www.cchem.berkeley.edu/jsngrp/fortran.html

    Parameters
    ----------
    sto: :class: `numpy.Array`
        Electrode stochiometry
    T: :class: `numpy.Array`
        Dimensional temperature
    T_inf: double
        Reference temperature
    E_D_s: double
        Solid diffusion activation energy
    R_g: double
        The ideal gas constant

    Returns
    -------
    : double
        Solid diffusivity
    """

    D_ref = 1 * 10 ** (-13)
    arrhenius = exp(E_D_s / R_g * (1 / T_inf - 1 / T))

    return D_ref * arrhenius



#
# Base class for convection submodels
#
import pybamm


class BaseModel(pybamm.BaseSubModel):
    """Base class for convection submodels.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel


    **Extends:** :class:`pybamm.BaseSubModel`
    """

    def __init__(self, param):
        super().__init__(param)

    def _get_standard_velocity_variables(self, v_box):
        """
        A private function to obtain the standard variables which
        can be derived from the fluid velocity.

        Parameters
        ----------
        v_box : :class:`pybamm.Symbol`
            The volume-averaged fluid velocity

        Returns
        -------
        variables : dict
            The variables which can be derived from the volume-averaged
            velocity.
        """

        vel_scale = self.param.velocity_scale

        # add more to this (x-averages etc)
        variables = {
            "Volume-averaged velocity": v_box,
            "Volume-averaged velocity [m.s-1]": vel_scale * v_box,
        }

        return variables

    def _get_standard_pressure_variables(self, p):
        """
        A private function to obtain the standard variables which
        can be derived from the pressure in the fluid.

        Parameters
        ----------
        p : :class:`pybamm.Symbol`
            The fluid pressure

        Returns
        -------
        variables : dict
            The variables which can be derived from the pressure.
        """

        # add more to this (x-averages etc)
        variables = {"Electrolyte pressure": p}

        return variables

    def _get_standard_vertical_velocity_variables(self, dVbox_dz):
        """
        A private function to obtain the standard variables which
        can be derived from the vertical velocity of the fluid.

        Parameters
        ----------
        dV_box_dz : :class:`pybamm.Symbol`
            The vertical velocity of the fluid

        Returns
        -------
        variables : dict
            The variables which can be derived from the vertical velocity.
        """
        vel_scale = self.param.velocity_scale
        L_z = self.param.L_z

        variables = {
            "Vertical volume-averaged acceleration": dVbox_dz,
            "Vertical volume-averaged acceleration [m.s-2]": vel_scale / L_z * dVbox_dz,
        }

        return variables

    def _separator_velocity(self, variables):
        """
        A private method to calculate x- and z-components of velocity in the separator

        Parameters
        ----------
        variables : dict
            Dictionary of variables in the whole model.

        Returns
        -------
        v_box_s : :class:`pybamm.Symbol`
            The x-component of velocity in the separator
        dVbox_dz : :class:`pybamm.Symbol`
            The z-component of velocity in the separator
        """
        # Set up
        param = self.param
        l_n = pybamm.geometric_parameters.l_n
        l_s = pybamm.geometric_parameters.l_s
        x_s = pybamm.standard_spatial_vars.x_s

        # Difference in negative and positive electrode velocities determines the
        # velocity in the separator
        i_boundary_cc = variables["Current collector current density"]
        v_box_n_right = param.beta_n * i_boundary_cc
        v_box_p_left = param.beta_p * i_boundary_cc
        d_vbox_s__dx = (v_box_p_left - v_box_n_right) / l_s

        # Simple formula for velocity in the separator
        dVbox_dz = pybamm.Concatenation(
            pybamm.FullBroadcast(
                0,
                "negative electrode",
                auxiliary_domains={"secondary": "current collector"},
            ),
            pybamm.PrimaryBroadcast(-d_vbox_s__dx, "separator"),
            pybamm.FullBroadcast(
                0,
                "positive electrode",
                auxiliary_domains={"secondary": "current collector"},
            ),
        )
        v_box_s = d_vbox_s__dx * (x_s - l_n) + v_box_n_right

        return v_box_s, dVbox_dz



#
# Class for lumped thermal submodel with no current collectors
#
import pybamm

from .base_x_lumped import BaseModel


class NoCurrentCollector(BaseModel):
    """
    Class for x-lumped thermal submodel without current collectors. Note: since
    there are no current collectors in this model, the electrochemical model
    must be 1D (x-direction only).

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel


    **Extends:** :class:`pybamm.thermal.BaseModel`
    """

    def __init__(self, param):
        super().__init__(param)

    def set_rhs(self, variables):
        T_av = variables["X-averaged cell temperature"]
        Q_av = variables["X-averaged total heating"]

        # Get effective properties
        rho_eff, _ = self._effective_properties()
        cooling_coeff = self._surface_cooling_coefficient()

        self.rhs = {
            T_av: (self.param.B * Q_av + cooling_coeff * T_av)
            / (self.param.C_th * rho_eff)
        }

    def _current_collector_heating(self, variables):
        """Returns zeros for current collector heat source terms"""
        Q_s_cn = pybamm.Scalar(0)
        Q_s_cp = pybamm.Scalar(0)
        return Q_s_cn, Q_s_cp

    def _surface_cooling_coefficient(self):
        """
        Returns the surface cooling coefficient in the absence of current
        collectors.
        """
        # Account for surface area to volume ratio in cooling coefficient
        # Note: assumes pouch cell geometry and volume doesn't include current
        # collectors
        A = self.param.l_y * self.param.l_z
        V = self.param.l_x * self.param.l_y * self.param.l_z
        return -2 * self.param.h * A / V / (self.param.delta ** 2)

    def _yz_average(self, var):
        """In 1D volume-averaged quantities are unchanged"""
        return var

    def _x_average(self, var, var_cn, var_cp):
        """
        Computes the X-average over the whole cell *not* including current
        collectors. This overwrites the default behaviour of 'base_thermal'.
        """
        return pybamm.x_average(var)

    def _effective_properties(self):
        """
        Computes the effective effective product of density and specific heat, and
        effective thermal conductivity, respectively. This overwrites the default
        behaviour of 'base_thermal'.
        """
        l_n = self.param.l_n
        l_s = self.param.l_s
        l_p = self.param.l_p
        rho_n = self.param.rho_n
        rho_s = self.param.rho_s
        rho_p = self.param.rho_p
        lambda_n = self.param.lambda_n
        lambda_s = self.param.lambda_s
        lambda_p = self.param.lambda_p

        rho_eff = (rho_n * l_n + rho_s * l_s + rho_p * l_p) / (l_n + l_n + l_p)
        lambda_eff = (lambda_n * l_n + lambda_s * l_s + lambda_p * l_p) / (
            l_n + l_n + l_p
        )

        return rho_eff, lambda_eff



from pybamm import exp


def nca_diffusivity_Kim2011(sto, T, T_inf, E_D_s, R_g):
    """
    NCA diffusivity as a function of stochiometry [1].

    References
    ----------
    .. [1] Kim, G. H., Smith, K., Lee, K. J., Santhanagopalan, S., & Pesaran, A.
    (2011). Multi-domain modeling of lithium-ion batteries encompassing
    multi-physics in varied length scales. Journal of The Electrochemical
    Society, 158(8), A955-A969.

    Parameters
    ----------
    sto: :class: `numpy.Array`
        Electrode stochiometry
    T: :class: `numpy.Array`
        Dimensional temperature
    T_inf: double
        Reference temperature
    E_D_s: double
        Solid diffusion activation energy
    R_g: double
        The ideal gas constant

    Returns
    -------
    : double
        Solid diffusivity
    """

    D_ref = 3 * 10 ** (-15)
    arrhenius = exp(E_D_s / R_g * (1 / T_inf - 1 / T))

    return D_ref * arrhenius



#
# Class for oxygen diffusion
#
import pybamm

from .full_oxygen_diffusion import Full, separator_and_positive_only


class Composite(Full):
    """Class for conservation of mass of oxygen. (Composite refers to composite
    expansion in asymptotic methods)
    In this model, extremely fast oxygen kinetics in the negative electrode imposes
    zero oxygen concentration there, and so the oxygen variable only lives in the
    separator and positive electrode. The boundary condition at the negative electrode/
    separator interface is homogeneous Dirichlet.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    reactions : dict
        Dictionary of reaction terms
    extended : bool
        Whether to include feedback from the first-order terms

    **Extends:** :class:`pybamm.oxygen_diffusion.Full`
    """

    def __init__(self, param, reactions, extended=False):
        super().__init__(param, reactions)
        self.extended = extended

    def get_coupled_variables(self, variables):

        tor_0 = separator_and_positive_only(
            variables["Leading-order electrolyte tortuosity"]
        )
        c_ox = variables["Separator and positive electrode oxygen concentration"]
        # TODO: allow charge and convection?
        # v_box = pybamm.Scalar(0)

        param = self.param

        N_ox_diffusion = -tor_0 * param.curlyD_ox * pybamm.grad(c_ox)

        N_ox = N_ox_diffusion  # + c_ox * v_box
        # Flux in the negative electrode is zero
        N_ox = pybamm.Concatenation(
            pybamm.FullBroadcast(0, "negative electrode", "current collector"), N_ox
        )

        variables.update(self._get_standard_flux_variables(N_ox))

        return variables

    def set_rhs(self, variables):
        "Composite reaction-diffusion with source terms from leading order"

        param = self.param

        eps_0 = separator_and_positive_only(variables["Leading-order porosity"])
        deps_0_dt = separator_and_positive_only(
            variables["Leading-order porosity change"]
        )
        c_ox = variables["Separator and positive electrode oxygen concentration"]
        N_ox = variables["Oxygen flux"].orphans[1]

        if self.extended is False:
            pos_reactions = sum(
                reaction["Positive"]["s_ox"]
                * variables["Leading-order " + reaction["Positive"]["aj"].lower()]
                for reaction in self.reactions.values()
            )
        else:
            pos_reactions = sum(
                reaction["Positive"]["s_ox"] * variables[reaction["Positive"]["aj"]]
                for reaction in self.reactions.values()
            )
        sep_reactions = pybamm.FullBroadcast(0, "separator", "current collector")
        source_terms_0 = (
            pybamm.Concatenation(sep_reactions, pos_reactions) / param.gamma_e
        )

        self.rhs = {
            c_ox: (1 / eps_0)
            * (-pybamm.div(N_ox) / param.C_e + source_terms_0 - c_ox * deps_0_dt)
        }



from pybamm import exp


def graphite_diffusivity_Kim2011(sto, T, T_inf, E_D_s, R_g):
    """
        Graphite diffusivity [1].

        References
        ----------
        .. [1] Kim, G. H., Smith, K., Lee, K. J., Santhanagopalan, S., & Pesaran, A.
        (2011). Multi-domain modeling of lithium-ion batteries encompassing
        multi-physics in varied length scales. Journal of The Electrochemical
        Society, 158(8), A955-A969.

        Parameters
        ----------
        sto: :class: `numpy.Array`
            Electrode stochiometry
        T: :class: `numpy.Array`
            Dimensional temperature
        T_inf: double
            Reference temperature
        E_D_s: double
            Solid diffusion activation energy
        R_g: double
            The ideal gas constant

        Returns
        -------
        : double
            Solid diffusivity
   """

    D_ref = 9 * 10 ** (-14)
    arrhenius = exp(E_D_s / R_g * (1 / T_inf - 1 / T))

    return D_ref * arrhenius



#
# Processed Variable class
#
import numbers
import numpy as np
import pybamm
import scipy.interpolate as interp


class ProcessedVariable(object):
    """
    An object that can be evaluated at arbitrary (scalars or vectors) t and x, and
    returns the (interpolated) value of the base variable at that t and x.

    Parameters
    ----------
    base_variable : :class:`pybamm.Symbol`
        A base variable with a method `evaluate(t,y)` that returns the value of that
        variable. Note that this can be any kind of node in the expression tree, not
        just a :class:`pybamm.Variable`.
        When evaluated, returns an array of size (m,n)
    solution : :class:`pybamm.Solution`
        The solution object to be used to create the processed variables
    interp_kind : str
        The method to use for interpolation
    known_evals : dict
        Dictionary of known evaluations, to be used to speed up finding the solution
    """

    def __init__(self, base_variable, solution, known_evals=None):
        self.base_variable = base_variable
        self.t_sol = solution.t
        self.u_sol = solution.y
        self.mesh = base_variable.mesh
        self.inputs = solution.inputs
        self.domain = base_variable.domain
        self.auxiliary_domains = base_variable.auxiliary_domains
        self.known_evals = known_evals

        if self.known_evals:
            self.base_eval, self.known_evals[solution.t[0]] = base_variable.evaluate(
                solution.t[0],
                solution.y[:, 0],
                {name: inp[0] for name, inp in solution.inputs.items()},
                known_evals=self.known_evals[solution.t[0]],
            )
        else:
            self.base_eval = base_variable.evaluate(
                solution.t[0],
                solution.y[:, 0],
                {name: inp[0] for name, inp in solution.inputs.items()},
            )

        # handle 2D (in space) finite element variables differently
        if (
            self.mesh
            and "current collector" in self.domain
            and isinstance(self.mesh[0], pybamm.ScikitSubMesh2D)
        ):
            if len(solution.t) == 1:
                # space only (steady solution)
                self.initialise_2Dspace_scikit_fem()
            else:
                self.initialise_3D_scikit_fem()

        # check variable shape
        else:
            if len(solution.t) == 1:
                raise pybamm.SolverError(
                    """
                    Solution time vector must have length > 1. Check whether simulation
                    terminated too early.
                    """
                )
            elif (
                isinstance(self.base_eval, numbers.Number)
                or len(self.base_eval.shape) == 0
                or self.base_eval.shape[0] == 1
            ):
                self.initialise_1D()
            else:
                n = self.mesh[0].npts
                base_shape = self.base_eval.shape[0]
                if base_shape in [n, n + 1]:
                    self.initialise_2D()
                else:
                    self.initialise_3D()

    def initialise_1D(self):
        # initialise empty array of the correct size
        entries = np.empty(len(self.t_sol))
        # Evaluate the base_variable index-by-index
        for idx in range(len(self.t_sol)):
            t = self.t_sol[idx]
            u = self.u_sol[:, idx]
            inputs = {name: inp[idx] for name, inp in self.inputs.items()}
            if self.known_evals:
                entries[idx], self.known_evals[t] = self.base_variable.evaluate(
                    t, u, inputs, known_evals=self.known_evals[t]
                )
            else:
                entries[idx] = self.base_variable.evaluate(t, u, inputs)

        # No discretisation provided, or variable has no domain (function of t only)
        self._interpolation_function = interp.interp1d(
            self.t_sol, entries, kind="linear", fill_value=np.nan, bounds_error=False
        )

        self.entries = entries
        self.dimensions = 1

    def initialise_2D(self):
        len_space = self.base_eval.shape[0]
        entries = np.empty((len_space, len(self.t_sol)))

        # Evaluate the base_variable index-by-index
        for idx in range(len(self.t_sol)):
            t = self.t_sol[idx]
            u = self.u_sol[:, idx]
            inputs = {name: inp[idx] for name, inp in self.inputs.items()}
            if self.known_evals:
                eval_and_known_evals = self.base_variable.evaluate(
                    t, u, inputs, known_evals=self.known_evals[t]
                )
                entries[:, idx] = eval_and_known_evals[0][:, 0]
                self.known_evals[t] = eval_and_known_evals[1]
            else:
                entries[:, idx] = self.base_variable.evaluate(t, u, inputs)[:, 0]

        # Process the discretisation to get x values
        nodes = self.mesh[0].nodes
        edges = self.mesh[0].edges
        if entries.shape[0] == len(nodes):
            space = nodes
        elif entries.shape[0] == len(edges):
            space = edges

        # add points outside domain for extrapolation to boundaries
        extrap_space_left = np.array([2 * space[0] - space[1]])
        extrap_space_right = np.array([2 * space[-1] - space[-2]])
        space = np.concatenate([extrap_space_left, space, extrap_space_right])
        extrap_entries_left = 2 * entries[0] - entries[1]
        extrap_entries_right = 2 * entries[-1] - entries[-2]
        entries_for_interp = np.vstack(
            [extrap_entries_left, entries, extrap_entries_right]
        )

        # assign attributes for reference (either x_sol or r_sol)
        self.entries = entries
        self.dimensions = 2
        if self.domain[0] in ["negative particle", "positive particle"]:
            self.first_dimension = "r"
            self.r_sol = space
        elif self.domain[0] in [
            "negative electrode",
            "separator",
            "positive electrode",
        ]:
            self.first_dimension = "x"
            self.x_sol = space
        elif self.domain == ["current collector"]:
            self.first_dimension = "z"
            self.z_sol = space
        else:
            self.first_dimension = "x"
            self.x_sol = space

        self.first_dim_pts = space

        # set up interpolation
        # note that the order of 't' and 'space' is the reverse of what you'd expect

        self._interpolation_function = interp.interp2d(
            self.t_sol, space, entries_for_interp, kind="linear", fill_value=np.nan
        )

    def initialise_3D(self):
        """
        Initialise a 3D object that depends on x and r, or x and z.
        """
        first_dim_nodes = self.mesh[0].nodes
        first_dim_edges = self.mesh[0].edges
        second_dim_pts = self.base_variable.secondary_mesh[0].nodes
        if self.base_eval.size // len(second_dim_pts) == len(first_dim_nodes):
            first_dim_pts = first_dim_nodes
        elif self.base_eval.size // len(second_dim_pts) == len(first_dim_edges):
            first_dim_pts = first_dim_edges

        # Process r-x or x-z
        if self.domain[0] in [
            "negative particle",
            "positive particle",
        ] and self.auxiliary_domains["secondary"][0] in [
            "negative electrode",
            "positive electrode",
        ]:
            self.first_dimension = "r"
            self.second_dimension = "x"
            self.r_sol = first_dim_pts
            self.x_sol = second_dim_pts
        elif self.domain[0] in [
            "negative electrode",
            "separator",
            "positive electrode",
        ] and self.auxiliary_domains["secondary"] == ["current collector"]:
            self.first_dimension = "x"
            self.second_dimension = "z"
            self.x_sol = first_dim_pts
            self.z_sol = second_dim_pts
        else:
            raise pybamm.DomainError(
                """ Cannot process 3D object with domain '{}'
                and auxiliary_domains '{}'""".format(
                    self.domain, self.auxiliary_domains
                )
            )

        first_dim_size = len(first_dim_pts)
        second_dim_size = len(second_dim_pts)
        entries = np.empty((first_dim_size, second_dim_size, len(self.t_sol)))

        # Evaluate the base_variable index-by-index
        for idx in range(len(self.t_sol)):
            t = self.t_sol[idx]
            u = self.u_sol[:, idx]
            inputs = {name: inp[idx] for name, inp in self.inputs.items()}
            if self.known_evals:
                eval_and_known_evals = self.base_variable.evaluate(
                    t, u, inputs, known_evals=self.known_evals[t]
                )
                entries[:, :, idx] = np.reshape(
                    eval_and_known_evals[0],
                    [first_dim_size, second_dim_size],
                    order="F",
                )
                self.known_evals[t] = eval_and_known_evals[1]
            else:
                entries[:, :, idx] = np.reshape(
                    self.base_variable.evaluate(t, u, inputs),
                    [first_dim_size, second_dim_size],
                    order="F",
                )

        # assign attributes for reference
        self.entries = entries
        self.dimensions = 3
        self.first_dim_pts = first_dim_pts
        self.second_dim_pts = second_dim_pts

        # set up interpolation
        self._interpolation_function = interp.RegularGridInterpolator(
            (first_dim_pts, second_dim_pts, self.t_sol),
            entries,
            method="linear",
            fill_value=np.nan,
        )

    def initialise_2Dspace_scikit_fem(self):
        y_sol = self.mesh[0].edges["y"]
        len_y = len(y_sol)
        z_sol = self.mesh[0].edges["z"]
        len_z = len(z_sol)

        # Evaluate the base_variable
        inputs = {name: inp[0] for name, inp in self.inputs.items()}

        entries = np.reshape(
            self.base_variable.evaluate(0, self.u_sol, inputs), [len_y, len_z]
        )

        # assign attributes for reference
        self.entries = entries
        self.dimensions = 2
        self.y_sol = y_sol
        self.z_sol = z_sol
        self.first_dimension = "y"
        self.second_dimension = "z"

        # set up interpolation
        self._interpolation_function = interp.interp2d(
            y_sol, z_sol, entries, kind="linear", fill_value=np.nan
        )

    def initialise_3D_scikit_fem(self):
        y_sol = self.mesh[0].edges["y"]
        len_y = len(y_sol)
        z_sol = self.mesh[0].edges["z"]
        len_z = len(z_sol)
        entries = np.empty((len_y, len_z, len(self.t_sol)))

        # Evaluate the base_variable index-by-index
        for idx in range(len(self.t_sol)):
            t = self.t_sol[idx]
            u = self.u_sol[:, idx]
            inputs = {name: inp[idx] for name, inp in self.inputs.items()}

            if self.known_evals:
                eval_and_known_evals = self.base_variable.evaluate(
                    t, u, inputs, known_evals=self.known_evals[t]
                )
                entries[:, :, idx] = np.reshape(eval_and_known_evals[0], [len_y, len_z])
                self.known_evals[t] = eval_and_known_evals[1]
            else:
                entries[:, :, idx] = np.reshape(
                    self.base_variable.evaluate(t, u, inputs), [len_y, len_z]
                )

        # assign attributes for reference
        self.entries = entries
        self.dimensions = 3
        self.y_sol = y_sol
        self.z_sol = z_sol
        self.first_dimension = "y"
        self.second_dimension = "z"

        # set up interpolation
        self._interpolation_function = interp.RegularGridInterpolator(
            (y_sol, z_sol, self.t_sol), entries, method="linear", fill_value=np.nan
        )

    def __call__(self, t=None, x=None, r=None, y=None, z=None, warn=True):
        """
        Evaluate the variable at arbitrary t (and x, r, y and/or z), using interpolation
        """
        if self.dimensions == 1:
            out = self._interpolation_function(t)
        elif self.dimensions == 2:
            if t is None:
                out = self._interpolation_function(y, z)
            else:
                out = self.call_2D(t, x, r, z)
        elif self.dimensions == 3:
            out = self.call_3D(t, x, r, y, z)
        if warn is True and np.isnan(out).any():
            pybamm.logger.warning(
                "Calling variable outside interpolation range (returns 'nan')"
            )
        return out

    def call_2D(self, t, x, r, z):
        "Evaluate a 2D variable"
        spatial_var = eval_dimension_name(self.first_dimension, x, r, None, z)
        return self._interpolation_function(t, spatial_var)

    def call_3D(self, t, x, r, y, z):
        "Evaluate a 3D variable"
        first_dim = eval_dimension_name(self.first_dimension, x, r, y, z)
        second_dim = eval_dimension_name(self.second_dimension, x, r, y, z)
        if isinstance(first_dim, np.ndarray):
            if isinstance(second_dim, np.ndarray) and isinstance(t, np.ndarray):
                first_dim = first_dim[:, np.newaxis, np.newaxis]
                second_dim = second_dim[:, np.newaxis]
            elif isinstance(second_dim, np.ndarray) or isinstance(t, np.ndarray):
                first_dim = first_dim[:, np.newaxis]
        else:
            if isinstance(second_dim, np.ndarray) and isinstance(t, np.ndarray):
                second_dim = second_dim[:, np.newaxis]

        return self._interpolation_function((first_dim, second_dim, t))

    @property
    def data(self):
        "Same as entries, but different name"
        return self.entries


def eval_dimension_name(name, x, r, y, z):
    if name == "x":
        out = x
    elif name == "r":
        out = r
    elif name == "y":
        out = y
    elif name == "z":
        out = z

    if out is None:
        raise ValueError("inputs {} cannot be None".format(name))
    else:
        return out



#
# Class for quick plotting of variables from models
#
import numpy as np
import pybamm
import warnings
from collections import defaultdict


def ax_min(data):
    "Calculate appropriate minimum axis value for plotting"
    data_min = np.nanmin(data)
    if data_min <= 0:
        return 1.04 * data_min
    else:
        return 0.96 * data_min


def ax_max(data):
    "Calculate appropriate maximum axis value for plotting"
    data_max = np.nanmax(data)
    if data_max <= 0:
        return 0.96 * data_max
    else:
        return 1.04 * data_max


def split_long_string(title, max_words=4):
    "Get title in a nice format"
    words = title.split()
    # Don't split if fits on one line, don't split just for units
    if len(words) <= max_words or words[max_words].startswith("["):
        return title
    else:
        first_line = (" ").join(words[:max_words])
        second_line = (" ").join(words[max_words:])
        return first_line + "\n" + second_line


class QuickPlot(object):
    """
    Generates a quick plot of a subset of key outputs of the model so that the model
    outputs can be easily assessed. The axis limits can be set using:
        self.axis["Variable name"] = [x_min, x_max, y_min, y_max]
    They can be reset to the default values by using self.reset_axis.

    Parameters
    ----------
    models: (iter of) :class:`pybamm.BaseModel`
        The model(s) to plot the outputs of.
    meshes: (iter of) :class:`pybamm.Mesh`
        The mesh(es) on which the model(s) were solved.
    solutions: (iter of) :class:`pybamm.Solver`
        The numerical solution(s) for the model(s) which contained the solution to the
        model(s).
    output_variables : list of str, optional
        List of variables to plot
    labels : list of str, optional
        Labels for the different models. Defaults to model names
    colors : list of str, optional
        The colors to loop over when plotting. Defaults to
        ["r", "b", "k", "g", "m", "c"]
    linestyles : list of str, optional
        The linestyles to loop over when plotting. Defaults to ["-", ":", "--", "-."]
    """

    def __init__(
        self,
        solutions,
        output_variables=None,
        labels=None,
        colors=None,
        linestyles=None,
    ):
        if isinstance(solutions, pybamm.Solution):
            solutions = [solutions]
        elif not isinstance(solutions, list):
            raise TypeError("'solutions' must be 'pybamm.Solution' or list")

        models = [solution.model for solution in solutions]

        # Set labels
        self.labels = labels or [model.name for model in models]

        # Set colors and linestyles
        self.colors = colors
        self.linestyles = linestyles

        # Time scale in hours
        self.time_scale = models[0].timescale_eval / 3600
        # Spatial scales (default to 1 if information not in model)
        variables = models[0].variables
        self.spatial_scales = {"x": 1, "y": 1, "z": 1, "r_n": 1, "r_p": 1}
        if "x [m]" and "x" in variables:
            self.spatial_scales["x"] = (variables["x [m]"] / variables["x"]).evaluate()[
                -1
            ]
        if "y [m]" and "y" in variables:
            self.spatial_scales["y"] = (variables["y [m]"] / variables["y"]).evaluate()[
                -1
            ]
        if "z [m]" and "z" in variables:
            self.spatial_scales["z"] = (variables["z [m]"] / variables["z"]).evaluate()[
                -1
            ]
        if "r_n [m]" and "r_n" in variables:
            self.spatial_scales["r_n"] = (
                variables["r_n [m]"] / variables["r_n"]
            ).evaluate()[-1]
        if "r_p [m]" and "r_p" in variables:
            self.spatial_scales["r_p"] = (
                variables["r_p [m]"] / variables["r_p"]
            ).evaluate()[-1]

        # Time parameters
        self.ts = [solution.t for solution in solutions]
        self.min_t = np.min([t[0] for t in self.ts]) * self.time_scale
        self.max_t = np.max([t[-1] for t in self.ts]) * self.time_scale

        # Default output variables for lead-acid and lithium-ion
        if output_variables is None:
            if isinstance(models[0], pybamm.lithium_ion.BaseModel):
                output_variables = [
                    "Negative particle surface concentration",
                    "Electrolyte concentration",
                    "Positive particle surface concentration",
                    "Current [A]",
                    "Negative electrode potential [V]",
                    "Electrolyte potential [V]",
                    "Positive electrode potential [V]",
                    "Terminal voltage [V]",
                ]
            elif isinstance(models[0], pybamm.lead_acid.BaseModel):
                output_variables = [
                    "Interfacial current density [A.m-2]",
                    "Electrolyte concentration [mol.m-3]",
                    "Current [A]",
                    "Porosity",
                    "Electrolyte potential [V]",
                    "Terminal voltage [V]",
                ]
            # else plot all variables in first model
            else:
                output_variables = models[0].variables

        self.set_output_variables(output_variables, solutions)
        self.reset_axis()

    def set_output_variables(self, output_variables, solutions):
        # Set up output variables
        self.variables = {}
        self.spatial_variable = {}

        # Calculate subplot positions based on number of variables supplied
        self.subplot_positions = {}
        self.n_rows = int(len(output_variables) // np.sqrt(len(output_variables)))
        self.n_cols = int(np.ceil(len(output_variables) / self.n_rows))

        # Process output variables into a form that can be plotted
        processed_variables = {}
        for solution in solutions:
            processed_variables[solution] = {}
            for variable_list in output_variables:
                # Make sure we always have a list of lists of variables
                if isinstance(variable_list, str):
                    variable_list = [variable_list]
                # Add all variables to the list of variables that should be processed
                processed_variables[solution].update(
                    {var: solution[var] for var in variable_list}
                )

        # Prepare dictionary of variables
        for k, variable_list in enumerate(output_variables):
            # Make sure we always have a list of lists of variables
            if isinstance(variable_list, str):
                variable_list = [variable_list]

            # Prepare list of variables
            key = tuple(variable_list)
            self.variables[key] = [None] * len(solutions)

            # process each variable in variable_list for each model
            for i, solution in enumerate(solutions):
                # self.variables is a dictionary of lists of lists
                self.variables[key][i] = [
                    processed_variables[solution][var] for var in variable_list
                ]

            # Make sure variables have the same dimensions and domain
            first_variable = self.variables[key][0][0]
            domain = first_variable.domain
            for variable in self.variables[key][0]:
                if variable.domain != domain:
                    raise ValueError("mismatching variable domains")

            # Set the x variable for any two-dimensional variables
            if first_variable.dimensions == 2:
                spatial_variable_key = first_variable.first_dimension
                spatial_variable_value = first_variable.first_dim_pts
                self.spatial_variable[key] = (
                    spatial_variable_key,
                    spatial_variable_value,
                )

            # Don't allow 3D variables
            elif any(var.dimensions == 3 for var in self.variables[key][0]):
                raise NotImplementedError("cannot plot 3D variables")

            # Define subplot position
            self.subplot_positions[key] = (self.n_rows, self.n_cols, k + 1)

    def reset_axis(self):
        """
        Reset the axis limits to the default values.
        These are calculated to fit around the minimum and maximum values of all the
        variables in each subplot
        """
        self.axis = {}
        for key, variable_lists in self.variables.items():
            if variable_lists[0][0].dimensions == 1:
                spatial_var_name, spatial_var_value = "x", None
                x_min = self.min_t
                x_max = self.max_t
            elif variable_lists[0][0].dimensions == 2:
                spatial_var_name, spatial_var_value = self.spatial_variable[key]
                if spatial_var_name == "r":
                    if "negative" in key[0].lower():
                        spatial_var_scaled = (
                            spatial_var_value * self.spatial_scales["r_n"]
                        )
                    elif "positive" in key[0].lower():
                        spatial_var_scaled = (
                            spatial_var_value * self.spatial_scales["r_p"]
                        )
                else:
                    spatial_var_scaled = (
                        spatial_var_value * self.spatial_scales[spatial_var_name]
                    )
                x_min = spatial_var_scaled[0]
                x_max = spatial_var_scaled[-1]

            # Get min and max y values
            y_min = np.min(
                [
                    ax_min(
                        var(
                            self.ts[i],
                            **{spatial_var_name: spatial_var_value},
                            warn=False
                        )
                    )
                    for i, variable_list in enumerate(variable_lists)
                    for var in variable_list
                ]
            )
            y_max = np.max(
                [
                    ax_max(
                        var(
                            self.ts[i],
                            **{spatial_var_name: spatial_var_value},
                            warn=False
                        )
                    )
                    for i, variable_list in enumerate(variable_lists)
                    for var in variable_list
                ]
            )
            if y_min == y_max:
                y_min -= 1
                y_max += 1
            self.axis[key] = [x_min, x_max, y_min, y_max]

    def plot(self, t):
        """Produces a quick plot with the internal states at time t.

        Parameters
        ----------
        t : float
            Dimensional time (in hours) at which to plot.
        """

        import matplotlib.pyplot as plt

        t /= self.time_scale
        self.fig, self.ax = plt.subplots(self.n_rows, self.n_cols, figsize=(15, 8))
        plt.tight_layout()
        plt.subplots_adjust(left=-0.1)
        self.plots = {}
        self.time_lines = {}

        colors = self.colors or ["r", "b", "k", "g", "m", "c"]
        linestyles = self.linestyles or ["-", ":", "--", "-."]
        fontsize = 42 // self.n_cols

        for k, (key, variable_lists) in enumerate(self.variables.items()):
            if len(self.variables) == 1:
                ax = self.ax
            else:
                ax = self.ax.flat[k]
            ax.set_xlim(self.axis[key][:2])
            ax.set_ylim(self.axis[key][2:])
            ax.xaxis.set_major_locator(plt.MaxNLocator(3))
            self.plots[key] = defaultdict(dict)
            # Set labels for the first subplot only (avoid repetition)
            if variable_lists[0][0].dimensions == 2:
                # 2D plot: plot as a function of x at time t
                spatial_var_name, spatial_var_value = self.spatial_variable[key]
                ax.set_xlabel(spatial_var_name + " [m]", fontsize=fontsize)
                for i, variable_list in enumerate(variable_lists):
                    for j, variable in enumerate(variable_list):
                        if spatial_var_name == "r":
                            if "negative" in key[0].lower():
                                spatial_scale = self.spatial_scales["r_n"]
                            elif "positive" in key[0].lower():
                                spatial_scale = self.spatial_scales["r_p"]
                        else:
                            spatial_scale = self.spatial_scales[spatial_var_name]
                        (self.plots[key][i][j],) = ax.plot(
                            spatial_var_value * spatial_scale,
                            variable(
                                t, **{spatial_var_name: spatial_var_value}, warn=False
                            ),
                            lw=2,
                            color=colors[i],
                            linestyle=linestyles[j],
                        )
            else:
                # 1D plot: plot as a function of time, indicating time t with a line
                ax.set_xlabel("Time [h]", fontsize=fontsize)
                for i, variable_list in enumerate(variable_lists):
                    for j, variable in enumerate(variable_list):
                        full_t = self.ts[i]
                        (self.plots[key][i][j],) = ax.plot(
                            full_t * self.time_scale,
                            variable(full_t, warn=False),
                            lw=2,
                            color=colors[i],
                            linestyle=linestyles[j],
                        )
                y_min, y_max = self.axis[key][2:]
                (self.time_lines[key],) = ax.plot(
                    [t * self.time_scale, t * self.time_scale], [y_min, y_max], "k--"
                )
            # Set either y label or legend entries
            if len(key) == 1:
                title = split_long_string(key[0])
                ax.set_title(title, fontsize=fontsize)
            else:
                ax.legend(
                    [split_long_string(s, 6) for s in key],
                    bbox_to_anchor=(0.5, 1),
                    fontsize=8,
                    loc="lower center",
                )
            if k == len(self.variables) - 1:
                ax.legend(self.labels, loc="upper right", bbox_to_anchor=(1, -0.2))

    def dynamic_plot(self, testing=False):
        """
        Generate a dynamic plot with a slider to control the time. We recommend using
        ipywidgets instead of this function if you are using jupyter notebooks
        """

        import matplotlib.pyplot as plt
        from matplotlib.widgets import Slider

        # create an initial plot at time 0
        self.plot(0)

        axcolor = "lightgoldenrodyellow"
        axfreq = plt.axes([0.315, 0.02, 0.37, 0.03], facecolor=axcolor)
        self.sfreq = Slider(axfreq, "Time", 0, self.max_t, valinit=0)
        self.sfreq.on_changed(self.update)

        # ignore the warning about tight layout
        warnings.simplefilter("ignore")
        self.fig.tight_layout()
        warnings.simplefilter("always")

        if not testing:  # pragma: no cover
            plt.show()

    def update(self, val):
        """
        Update the plot in self.plot() with values at new time
        """
        t = self.sfreq.val
        t_dimensionless = t / self.time_scale
        for key, plot in self.plots.items():
            if self.variables[key][0][0].dimensions == 2:
                spatial_var_name, spatial_var_value = self.spatial_variable[key]
                for i, variable_lists in enumerate(self.variables[key]):
                    for j, variable in enumerate(variable_lists):
                        plot[i][j].set_ydata(
                            variable(
                                t_dimensionless,
                                **{spatial_var_name: spatial_var_value},
                                warn=False
                            )
                        )
            else:
                self.time_lines[key].set_xdata([t])

        self.fig.canvas.draw_idle()



#
# Function classes and methods
#
import autograd
import numbers
import numpy as np
import pybamm


class Function(pybamm.Symbol):
    """A node in the expression tree representing an arbitrary function

    Parameters
    ----------
    function : method
        A function can have 0 or many inputs. If no inputs are given, self.evaluate()
        simply returns func(). Otherwise, self.evaluate(t, y, u) returns
        func(child0.evaluate(t, y, u), child1.evaluate(t, y, u), etc).
    children : :class:`pybamm.Symbol`
        The children nodes to apply the function to
    derivative : str, optional
        Which derivative to use when differentiating ("autograd" or "derivative").
        Default is "autograd".
    differentiated_function : method, optional
        The function which was differentiated to obtain this one. Default is None.
    **Extends:** :class:`pybamm.Symbol`
    """

    def __init__(
        self,
        function,
        *children,
        name=None,
        derivative="autograd",
        differentiated_function=None
    ):
        # Turn numbers into scalars
        children = list(children)
        for idx, child in enumerate(children):
            if isinstance(child, numbers.Number):
                children[idx] = pybamm.Scalar(child)

        if name is not None:
            self.name = name
        else:
            try:
                name = "function ({})".format(function.__name__)
            except AttributeError:
                name = "function ({})".format(function.__class__)
        domain = self.get_children_domains(children)
        auxiliary_domains = self.get_children_auxiliary_domains(children)

        self.function = function
        self.derivative = derivative
        self.differentiated_function = differentiated_function

        super().__init__(
            name, children=children, domain=domain, auxiliary_domains=auxiliary_domains
        )

    def get_children_domains(self, children_list):
        """Obtains the unique domain of the children. If the
        children have different domains then raise an error"""

        domains = [child.domain for child in children_list if child.domain != []]

        # check that there is one common domain amongst children
        distinct_domains = set(tuple(dom) for dom in domains)

        if len(distinct_domains) > 1:
            raise pybamm.DomainError(
                "Functions can only be applied to variables on the same domain"
            )
        elif len(distinct_domains) == 0:
            domain = []
        else:
            domain = domains[0]

        return domain

    def diff(self, variable):
        """ See :meth:`pybamm.Symbol.diff()`. """
        if variable.id == self.id:
            return pybamm.Scalar(1)
        else:
            children = self.orphans
            partial_derivatives = [None] * len(children)
            for i, child in enumerate(self.children):
                # if variable appears in the function,use autograd to differentiate
                # function, and apply chain rule
                if variable.id in [symbol.id for symbol in child.pre_order()]:
                    partial_derivatives[i] = self._function_diff(
                        children, i
                    ) * child.diff(variable)

            # remove None entries
            partial_derivatives = list(filter(None, partial_derivatives))

            derivative = sum(partial_derivatives)
            if derivative == 0:
                derivative = pybamm.Scalar(0)

            return derivative

    def _function_diff(self, children, idx):
        """
        Derivative with respect to child number 'idx'.
        See :meth:`pybamm.Symbol._diff()`.
        """
        # Store differentiated function, needed in case we want to convert to CasADi
        if self.derivative == "autograd":
            return Function(
                autograd.elementwise_grad(self.function, idx),
                *children,
                differentiated_function=self.function
            )
        elif self.derivative == "derivative":
            if len(children) > 1:
                raise ValueError(
                    """
                    differentiation using '.derivative()' not implemented for functions
                    with more than one child
                    """
                )
            else:
                # keep using "derivative" as derivative
                return pybamm.Function(
                    self.function.derivative(),
                    *children,
                    derivative="derivative",
                    differentiated_function=self.function
                )

    def _function_jac(self, children_jacs):
        """ Calculate the jacobian of a function. """

        if all(child.evaluates_to_number() for child in self.children):
            jacobian = pybamm.Scalar(0)
        else:
            # if at least one child contains variable dependence, then
            # calculate the required partial jacobians and add them
            jacobian = None
            children = self.orphans
            for i, child in enumerate(children):
                if not child.evaluates_to_number():
                    jac_fun = self._function_diff(children, i) * children_jacs[i]
                    jac_fun.clear_domains()
                    if jacobian is None:
                        jacobian = jac_fun
                    else:
                        jacobian += jac_fun

        return jacobian

    def evaluate(self, t=None, y=None, u=None, known_evals=None):
        """ See :meth:`pybamm.Symbol.evaluate()`. """
        if known_evals is not None:
            if self.id not in known_evals:
                evaluated_children = [None] * len(self.children)
                for i, child in enumerate(self.children):
                    evaluated_children[i], known_evals = child.evaluate(
                        t, y, u, known_evals=known_evals
                    )
                known_evals[self.id] = self._function_evaluate(evaluated_children)
            return known_evals[self.id], known_evals
        else:
            evaluated_children = [child.evaluate(t, y, u) for child in self.children]
            return self._function_evaluate(evaluated_children)

    def _evaluate_for_shape(self):
        """
        Default behaviour: has same shape as all child
        See :meth:`pybamm.Symbol.evaluate_for_shape()`
        """
        evaluated_children = [child.evaluate_for_shape() for child in self.children]
        return self._function_evaluate(evaluated_children)

    def _function_evaluate(self, evaluated_children):
        return self.function(*evaluated_children)

    def new_copy(self):
        """ See :meth:`pybamm.Symbol.new_copy()`. """
        children_copy = [child.new_copy() for child in self.children]
        return self._function_new_copy(children_copy)

    def _function_new_copy(self, children):
        """Returns a new copy of the function.

        Inputs
        ------
        children : : list
            A list of the children of the function

        Returns
        -------
            : :pybamm.Function
            A new copy of the function
        """
        return pybamm.Function(
            self.function,
            *children,
            name=self.name,
            derivative=self.derivative,
            differentiated_function=self.differentiated_function
        )

    def _function_simplify(self, simplified_children):
        """
        Simplifies the function.

        Inputs
        ------
        simplified_children: : list
            A list of simplified children of the function

        Returns
        -------
         :: pybamm.Scalar() if no children
         :: pybamm.Function if there are children
        """
        return self._function_new_copy(simplified_children)


class SpecificFunction(Function):
    """
    Parent class for the specific functions, which implement their own `diff`
    operators directly.

    Parameters
    ----------
    function : method
        Function to be applied to child
    child : :class:`pybamm.Symbol`
        The child to apply the function to

    """

    def __init__(self, function, child):
        super().__init__(function, child)

    def _function_new_copy(self, children):
        """ See :meth:`pybamm.Function._function_new_copy()` """
        return self.__class__(*children)

    def _function_simplify(self, simplified_children):
        """ See :meth:`pybamm.Function._function_simplify()` """
        return self.__class__(*simplified_children)


class Arcsinh(SpecificFunction):
    """ Arcsinh function """

    def __init__(self, child):
        super().__init__(np.arcsinh, child)

    def _function_diff(self, children, idx):
        """ See :meth:`pybamm.Symbol._function_diff()`. """
        return 1 / Sqrt(children[0] ** 2 + 1)


def arcsinh(child):
    " Returns arcsinh function of child. "
    return pybamm.simplify_if_constant(Arcsinh(child), keep_domains=True)


class Cos(SpecificFunction):
    """ Cosine function """

    def __init__(self, child):
        super().__init__(np.cos, child)

    def _function_diff(self, children, idx):
        """ See :meth:`pybamm.Symbol._function_diff()`. """
        return -Sin(children[0])


def cos(child):
    " Returns cosine function of child. "
    return pybamm.simplify_if_constant(Cos(child), keep_domains=True)


class Cosh(SpecificFunction):
    """ Hyberbolic cosine function """

    def __init__(self, child):
        super().__init__(np.cosh, child)

    def _function_diff(self, children, idx):
        """ See :meth:`pybamm.Function._function_diff()`. """
        return Sinh(children[0])


def cosh(child):
    " Returns hyperbolic cosine function of child. "
    return pybamm.simplify_if_constant(Cosh(child), keep_domains=True)


class Exponential(SpecificFunction):
    """ Exponential function """

    def __init__(self, child):
        super().__init__(np.exp, child)

    def _function_diff(self, children, idx):
        """ See :meth:`pybamm.Function._function_diff()`. """
        return Exponential(children[0])


def exp(child):
    " Returns exponential function of child. "
    return pybamm.simplify_if_constant(Exponential(child), keep_domains=True)


class Log(SpecificFunction):
    """ Logarithmic function """

    def __init__(self, child):
        super().__init__(np.log, child)

    def _function_evaluate(self, evaluated_children):
        # don't raise RuntimeWarning for NaNs
        with np.errstate(invalid="ignore"):
            return np.log(*evaluated_children)

    def _function_diff(self, children, idx):
        """ See :meth:`pybamm.Function._function_diff()`. """
        return 1 / children[0]


def log(child, base="e"):
    " Returns logarithmic function of child (any base, default 'e'). "
    if base == "e":
        return pybamm.simplify_if_constant(Log(child), keep_domains=True)
    else:
        return Log(child) / np.log(base)


def log10(child):
    " Returns logarithmic function of child, with base 10. "
    return log(child, base=10)


def max(child):
    " Returns max function of child. "
    return pybamm.simplify_if_constant(Function(np.max, child), keep_domains=True)


def min(child):
    " Returns min function of child. "
    return pybamm.simplify_if_constant(Function(np.min, child), keep_domains=True)


def sech(child):
    " Returns hyperbolic sec function of child. "
    return pybamm.simplify_if_constant(1 / Cosh(child), keep_domains=True)


class Sin(SpecificFunction):
    """ Sine function """

    def __init__(self, child):
        super().__init__(np.sin, child)

    def _function_diff(self, children, idx):
        """ See :meth:`pybamm.Function._function_diff()`. """
        return Cos(children[0])


def sin(child):
    " Returns sine function of child. "
    return pybamm.simplify_if_constant(Sin(child), keep_domains=True)


class Sinh(SpecificFunction):
    """ Hyperbolic sine function """

    def __init__(self, child):
        super().__init__(np.sinh, child)

    def _function_diff(self, children, idx):
        """ See :meth:`pybamm.Function._function_diff()`. """
        return Cosh(children[0])


def sinh(child):
    " Returns hyperbolic sine function of child. "
    return pybamm.simplify_if_constant(Sinh(child), keep_domains=True)


class Sqrt(SpecificFunction):
    """ Square root function """

    def __init__(self, child):
        super().__init__(np.sqrt, child)

    def _function_evaluate(self, evaluated_children):
        # don't raise RuntimeWarning for NaNs
        with np.errstate(invalid="ignore"):
            return np.sqrt(*evaluated_children)

    def _function_diff(self, children, idx):
        """ See :meth:`pybamm.Function._function_diff()`. """
        return 1 / (2 * Sqrt(children[0]))


def sqrt(child):
    " Returns square root function of child. "
    return pybamm.simplify_if_constant(Sqrt(child), keep_domains=True)


class Tanh(SpecificFunction):
    """ Hyperbolic tan function """

    def __init__(self, child):
        super().__init__(np.tanh, child)

    def _function_diff(self, children, idx):
        """ See :meth:`pybamm.Function._function_diff()`. """
        return sech(children[0]) ** 2


def tanh(child):
    " Returns hyperbolic tan function of child. "
    return pybamm.simplify_if_constant(Tanh(child), keep_domains=True)



#
# Class for isothermal submodel
#
import pybamm

from ..base_thermal import BaseThermal


class Isothermal(BaseThermal):
    """Class for isothermal submodel.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel


    **Extends:** :class:`pybamm.thermal.BaseThermal`
    """

    def __init__(self, param):
        super().__init__(param)

    def get_fundamental_variables(self):

        T_x_av = pybamm.PrimaryBroadcast(self.param.T_init, "current collector")
        T_n = pybamm.PrimaryBroadcast(T_x_av, "negative electrode")
        T_s = pybamm.PrimaryBroadcast(T_x_av, "separator")
        T_p = pybamm.PrimaryBroadcast(T_x_av, "positive electrode")
        T = pybamm.Concatenation(T_n, T_s, T_p)

        T_cn = T_x_av
        T_cp = T_x_av

        variables = self._get_standard_fundamental_variables(T, T_cn, T_cp)
        return variables

    def get_coupled_variables(self, variables):
        variables.update(
            {
                "Ohmic heating": pybamm.Scalar(0),
                "Ohmic heating [W.m-3]": pybamm.Scalar(0),
                "Irreversible electrochemical heating": pybamm.Scalar(0),
                "Irreversible electrochemical heating [W.m-3]": pybamm.Scalar(0),
                "Reversible heating": pybamm.Scalar(0),
                "Reversible heating [W.m-3]": pybamm.Scalar(0),
                "Total heating": pybamm.Scalar(0),
                "Total heating [W.m-3]": pybamm.Scalar(0),
                "X-averaged total heating": pybamm.Scalar(0),
                "X-averaged total heating [W.m-3]": pybamm.Scalar(0),
                "Volume-averaged total heating": pybamm.Scalar(0),
                "Volume-averaged total heating [W.m-3]": pybamm.Scalar(0),
            }
        )
        return variables

    def _flux_law(self, T):
        """Zero heat flux since temperature is constant"""
        q = pybamm.FullBroadcast(
            pybamm.Scalar(0),
            ["negative electrode", "separator", "positive electrode"],
            "current collector",
        )
        return q

    def _yz_average(self, var):
        """
        Temperature is uniform and heat source terms are zero, so the average
        returns the input variable.
        """
        return var

    def _x_average(self, var, var_cn, var_cp):
        """
        Temperature is uniform and heat source terms are zero, so the average
        returns zeros broadcasted onto the current collector domain.
        """
        return pybamm.PrimaryBroadcast(0, "current collector")



#
# Parameter sets from papers
#
"""
Parameter sets from papers. The 'citation' entry provides a reference to the appropriate
paper in the file "pybamm/CITATIONS.txt". To see which parameter sets have been used in
your simulation, add the line "pybamm.print_citations()" to your script.
"""

#
# Lithium-ion
#
Marquis2019 = {
    "chemistry": "lithium-ion",
    "cell": "kokam_Marquis2019",
    "anode": "graphite_mcmb2528_Marquis2019",
    "separator": "separator_Marquis2019",
    "cathode": "lico2_Marquis2019",
    "electrolyte": "lipf6_Marquis2019",
    "experiment": "1C_discharge_from_full_Marquis2019",
    "citation": "marquis2019asymptotic",
}

NCA_Kim2011 = {
    "chemistry": "lithium-ion",
    "cell": "Kim2011",
    "anode": "graphite_Kim2011",
    "separator": "separator_Kim2011",
    "cathode": "nca_Kim2011",
    "electrolyte": "lipf6_Kim2011",
    "experiment": "1C_discharge_from_full_Kim2011",
    "citation": "kim2011multi",
}

Chen2020 = {
    "chemistry": "lithium-ion",
    "cell": "LGM50_Chen2020",
    "anode": "graphite_Chen2020",
    "separator": "separator_Chen2020",
    "cathode": "nmc_Chen2020",
    "electrolyte": "lipf6_Nyman2008",
    "experiment": "1C_discharge_from_full_Chen2020",
    "citation": "Chen2020",
}
#
# Lead-acid
#
Sulzer2019 = {
    "chemistry": "lead-acid",
    "cell": "BBOXX_Sulzer2019",
    "anode": "lead_Sulzer2019",
    "separator": "agm_Sulzer2019",
    "cathode": "lead_dioxide_Sulzer2019",
    "electrolyte": "sulfuric_acid_Sulzer2019",
    "experiment": "1C_discharge_from_full",
    "citation": "sulzer2019physical",
}



#
# Base solver class
#
import casadi
import pybamm
import numbers
import numpy as np
from scipy import optimize
from scipy.sparse import issparse
import sys
import itertools


class BaseSolver(object):
    """Solve a discretised model.

    Parameters
    ----------
    method : str, optional
        The method to use for integration, specific to each solver
    rtol : float, optional
        The relative tolerance for the solver (default is 1e-6).
    atol : float, optional
        The absolute tolerance for the solver (default is 1e-6).
    root_method : str, optional
        The method to use to find initial conditions (default is "casadi"). If "casadi",
        the solver uses casadi's Newton rootfinding algorithm to find initial
        conditions. Otherwise, the solver uses 'scipy.optimize.root' with method
        specified by 'root_method' (e.g. "lm", "hybr", ...)
    root_tol : float, optional
        The tolerance for the initial-condition solver (default is 1e-6).
    max_steps: int, optional
        The maximum number of steps the solver will take before terminating
        (default is 1000).
    """

    def __init__(
        self,
        method=None,
        rtol=1e-6,
        atol=1e-6,
        root_method="casadi",
        root_tol=1e-6,
        max_steps=1000,
    ):
        self._method = method
        self._rtol = rtol
        self._atol = atol
        self.root_method = root_method
        self.root_tol = root_tol
        self.max_steps = max_steps

        self.models_set_up = set()

        # Defaults, can be overwritten by specific solver
        self.name = "Base solver"
        self.ode_solver = False

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, value):
        self._method = value

    @property
    def rtol(self):
        return self._rtol

    @rtol.setter
    def rtol(self, value):
        self._rtol = value

    @property
    def atol(self):
        return self._atol

    @atol.setter
    def atol(self, value):
        self._atol = value

    @property
    def root_method(self):
        return self._root_method

    @root_method.setter
    def root_method(self, method):
        self._root_method = method

    @property
    def root_tol(self):
        return self._root_tol

    @root_tol.setter
    def root_tol(self, tol):
        self._root_tol = tol

    @property
    def max_steps(self):
        return self._max_steps

    @max_steps.setter
    def max_steps(self, max_steps):
        self._max_steps = max_steps

    def set_up(self, model, inputs=None):
        """Unpack model, perform checks, simplify and calculate jacobian.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate. Must have attributes rhs and
            initial_conditions
        inputs : dict, optional
            Any input parameters to pass to the model when solving

        """
        inputs = inputs or {}
        y0 = model.concatenated_initial_conditions.evaluate(0, None, inputs)

        # Set model timescale
        model.timescale_eval = model.timescale.evaluate(u=inputs)

        # Check model.algebraic for ode solvers
        if self.ode_solver is True and len(model.algebraic) > 0:
            raise pybamm.SolverError(
                "Cannot use ODE solver '{}' to solve DAE model".format(self.name)
            )

        if self.ode_solver is True:
            self.root_method = None
        if (
            isinstance(self, pybamm.CasadiSolver) or self.root_method == "casadi"
        ) and model.convert_to_format != "casadi":
            pybamm.logger.warning(
                f"Converting {model.name} to CasADi for solving with CasADi solver"
            )
            model.convert_to_format = "casadi"

        if model.convert_to_format != "casadi":
            simp = pybamm.Simplification()
            # Create Jacobian from concatenated rhs and algebraic
            y = pybamm.StateVector(slice(0, np.size(y0)))
            # set up Jacobian object, for re-use of dict
            jacobian = pybamm.Jacobian()
        else:
            # Convert model attributes to casadi
            t_casadi = casadi.MX.sym("t")
            y_diff = casadi.MX.sym(
                "y_diff", len(model.concatenated_rhs.evaluate(0, y0, inputs))
            )
            y_alg = casadi.MX.sym(
                "y_alg", len(model.concatenated_algebraic.evaluate(0, y0, inputs))
            )
            y_casadi = casadi.vertcat(y_diff, y_alg)
            u_casadi = {}
            for name, value in inputs.items():
                if isinstance(value, numbers.Number):
                    u_casadi[name] = casadi.MX.sym(name)
                else:
                    u_casadi[name] = casadi.MX.sym(name, value.shape[0])
            u_casadi_stacked = casadi.vertcat(*[u for u in u_casadi.values()])

        def process(func, name, use_jacobian=None):
            def report(string):
                # don't log event conversion
                if "event" not in string:
                    pybamm.logger.info(string)

            if use_jacobian is None:
                use_jacobian = model.use_jacobian
            if model.convert_to_format != "casadi":
                # Process with pybamm functions
                if model.use_simplify:
                    report(f"Simplifying {name}")
                    func = simp.simplify(func)
                if use_jacobian:
                    report(f"Calculating jacobian for {name}")
                    jac = jacobian.jac(func, y)
                    if model.use_simplify:
                        report(f"Simplifying jacobian for {name}")
                        jac = simp.simplify(jac)
                    if model.convert_to_format == "python":
                        report(f"Converting jacobian for {name} to python")
                        jac = pybamm.EvaluatorPython(jac)
                    jac = jac.evaluate
                else:
                    jac = None
                if model.convert_to_format == "python":
                    report(f"Converting {name} to python")
                    func = pybamm.EvaluatorPython(func)
                func = func.evaluate
            else:
                # Process with CasADi
                report(f"Converting {name} to CasADi")
                func = func.to_casadi(t_casadi, y_casadi, u_casadi)
                if use_jacobian:
                    report(f"Calculating jacobian for {name} using CasADi")
                    jac_casadi = casadi.jacobian(func, y_casadi)
                    jac = casadi.Function(
                        name, [t_casadi, y_casadi, u_casadi_stacked], [jac_casadi]
                    )
                else:
                    jac = None
                func = casadi.Function(
                    name, [t_casadi, y_casadi, u_casadi_stacked], [func]
                )
            if name == "residuals":
                func_call = Residuals(func, name, model)
            else:
                func_call = SolverCallable(func, name, model)
            func_call.set_inputs(inputs)
            if jac is not None:
                jac_call = SolverCallable(jac, name + "_jac", model)
                jac_call.set_inputs(inputs)
            else:
                jac_call = None
            return func, func_call, jac_call

        # Check for heaviside functions in rhs and algebraic and add discontinuity
        # events if these exist.
        # Note: only checks for the case of t < X, t <= X, X < t, or X <= t, but also
        # accounts for the fact that t might be dimensional
        # Only do this for DAE models as ODE models can deal with discontinuities fine
        if len(model.algebraic) > 0:
            for symbol in itertools.chain(
                model.concatenated_rhs.pre_order(),
                model.concatenated_algebraic.pre_order(),
            ):
                if isinstance(symbol, pybamm.Heaviside):
                    # Dimensionless
                    if symbol.right.id == pybamm.t.id:
                        expr = symbol.left
                    elif symbol.left.id == pybamm.t.id:
                        expr = symbol.right
                    # Dimensional
                    elif symbol.right.id == (pybamm.t * model.timescale).id:
                        expr = symbol.left.new_copy() / symbol.right.right.new_copy()
                    elif symbol.left.id == (pybamm.t * model.timescale).id:
                        expr = symbol.right.new_copy() / symbol.left.right.new_copy()

                    model.events.append(
                        pybamm.Event(
                            str(symbol), expr.new_copy(), pybamm.EventType.DISCONTINUITY
                        )
                    )

        # Process rhs, algebraic and event expressions
        rhs, rhs_eval, jac_rhs = process(model.concatenated_rhs, "RHS")
        algebraic, algebraic_eval, jac_algebraic = process(
            model.concatenated_algebraic, "algebraic"
        )
        terminate_events_eval = [
            process(event.expression, "event", use_jacobian=False)[1]
            for event in model.events
            if event.event_type == pybamm.EventType.TERMINATION
        ]

        # discontinuity events are evaluated before the solver is called, so don't need
        # to process them
        discontinuity_events_eval = [
            event
            for event in model.events
            if event.event_type == pybamm.EventType.DISCONTINUITY
        ]

        # Add the solver attributes
        model.rhs_eval = rhs_eval
        model.algebraic_eval = algebraic_eval
        model.jac_algebraic_eval = jac_algebraic
        model.terminate_events_eval = terminate_events_eval
        model.discontinuity_events_eval = discontinuity_events_eval

        # Save CasADi functions for the CasADi solver
        # Note: when we pass to casadi the ode part of the problem must be in explicit
        # form so we pre-multiply by the inverse of the mass matrix
        if self.root_method == "casadi" or isinstance(self, pybamm.CasadiSolver):
            mass_matrix_inv = casadi.MX(model.mass_matrix_inv.entries)
            explicit_rhs = mass_matrix_inv @ rhs(t_casadi, y_casadi, u_casadi_stacked)
            model.casadi_rhs = casadi.Function(
                "rhs", [t_casadi, y_casadi, u_casadi_stacked], [explicit_rhs]
            )
            model.casadi_algebraic = algebraic
        # Calculate consistent initial conditions for the algebraic equations
        if len(model.algebraic) > 0:
            all_states = pybamm.NumpyConcatenation(
                model.concatenated_rhs, model.concatenated_algebraic
            )
            # Process again, uses caching so should be quick
            residuals, residuals_eval, jacobian_eval = process(all_states, "residuals")
            model.residuals_eval = residuals_eval
            model.jacobian_eval = jacobian_eval
            y0_guess = y0.flatten()
            model.y0 = self.calculate_consistent_state(model, 0, y0_guess, inputs)
        else:
            # can use DAE solver to solve ODE model
            model.residuals_eval = Residuals(rhs, "residuals", model)
            model.jacobian_eval = jac_rhs
            model.y0 = y0.flatten()

        pybamm.logger.info("Finish solver set-up")

    def set_inputs(self, model, ext_and_inputs):
        """
        Set values that are controlled externally, such as external variables and input
        parameters

        Parameters
        ----------
        ext_and_inputs : dict
            Any external variables or input parameters to pass to the model when solving

        """
        model.rhs_eval.set_inputs(ext_and_inputs)
        model.algebraic_eval.set_inputs(ext_and_inputs)
        model.residuals_eval.set_inputs(ext_and_inputs)
        for evnt in model.terminate_events_eval:
            evnt.set_inputs(ext_and_inputs)
        if model.jacobian_eval:
            model.jacobian_eval.set_inputs(ext_and_inputs)

    def calculate_consistent_state(self, model, time=0, y0_guess=None, inputs=None):
        """
        Calculate consistent state for the algebraic equations through
        root-finding

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model for which to calculate initial conditions.
        time : float
            The time at which to calculate the states
        y0_guess : :class:`np.array`
            Guess for the rootfinding
        inputs : dict, optional
            Any input parameters to pass to the model when solving

        Returns
        -------
        y0_consistent : array-like, same shape as y0_guess
            Initial conditions that are consistent with the algebraic equations (roots
            of the algebraic equations)
        """
        pybamm.logger.info("Start calculating consistent states")
        if y0_guess is None:
            y0_guess = model.concatenated_initial_conditions.flatten()

        # Split y0_guess into differential and algebraic
        len_rhs = model.rhs_eval(time, y0_guess).shape[0]
        y0_diff, y0_alg_guess = np.split(y0_guess, [len_rhs])
        inputs = inputs or {}

        # Solve using casadi or scipy
        if self.root_method == "casadi":
            # Set up
            u_stacked = casadi.vertcat(*[x for x in inputs.values()])
            u = casadi.MX.sym("u", u_stacked.shape[0])
            y_alg = casadi.MX.sym("y_alg", y0_alg_guess.shape[0])
            y = casadi.vertcat(y0_diff, y_alg)
            alg_root = model.casadi_algebraic(time, y, u)
            # Solve
            # set error_on_fail to False and just check the final output is small
            # enough
            roots = casadi.rootfinder(
                "roots",
                "newton",
                dict(x=y_alg, p=u, g=alg_root),
                {"abstol": self.root_tol},
            )
            try:
                y0_alg = roots(y0_alg_guess, u_stacked).full().flatten()
                success = True
                message = None
                # Check final output
                fun = model.casadi_algebraic(
                    time, casadi.vertcat(y0_diff, y0_alg), u_stacked
                )
            except RuntimeError as err:
                success = False
                message = err.args[0]
                fun = None
        else:
            algebraic = model.algebraic_eval
            jac = model.jac_algebraic_eval

            def root_fun(y0_alg):
                "Evaluates algebraic using y0_diff (fixed) and y0_alg (changed by algo)"
                y0 = np.concatenate([y0_diff, y0_alg])
                out = algebraic(time, y0)
                pybamm.logger.debug(
                    "Evaluating algebraic equations at t={}, L2-norm is {}".format(
                        time * model.timescale, np.linalg.norm(out)
                    )
                )
                return out

            if jac:
                if issparse(jac(0, y0_guess)):

                    def jac_fn(y0_alg):
                        """
                        Evaluates jacobian using y0_diff (fixed) and y0_alg (varying)
                        """
                        y0 = np.concatenate([y0_diff, y0_alg])
                        return jac(0, y0)[:, len_rhs:].toarray()

                else:

                    def jac_fn(y0_alg):
                        """
                        Evaluates jacobian using y0_diff (fixed) and y0_alg (varying)
                        """
                        y0 = np.concatenate([y0_diff, y0_alg])
                        return jac(0, y0)[:, len_rhs:]

            else:
                jac_fn = None
            # Find the values of y0_alg that are roots of the algebraic equations
            sol = optimize.root(
                root_fun,
                y0_alg_guess,
                jac=jac_fn,
                method=self.root_method,
                tol=self.root_tol,
            )
            pybamm.citations.register("virtanen2020scipy")

            # Set outputs
            y0_alg = sol.x
            success = sol.success
            fun = sol.fun
            message = sol.message

        if success and np.all(fun < self.root_tol * len(y0_alg)):
            # Return full set of consistent initial conditions (y0_diff unchanged)
            y0_consistent = np.concatenate([y0_diff, y0_alg])
            pybamm.logger.info("Finish calculating consistent initial conditions")
            return y0_consistent
        elif not success:
            raise pybamm.SolverError(
                "Could not find consistent initial conditions: {}".format(message)
            )
        else:
            raise pybamm.SolverError(
                """
                Could not find consistent initial conditions: solver terminated
                successfully, but maximum solution error ({}) above tolerance ({})
                """.format(
                    np.max(fun), self.root_tol * len(y0_alg)
                )
            )

    def solve(self, model, t_eval, external_variables=None, inputs=None):
        """
        Execute the solver setup and calculate the solution of the model at
        specified times.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate. Must have attributes rhs and
            initial_conditions
        t_eval : numeric type
            The times at which to compute the solution
        external_variables : dict
            A dictionary of external variables and their corresponding
            values at the current time
        inputs : dict, optional
            Any input parameters to pass to the model when solving

        Raises
        ------
        :class:`pybamm.ModelError`
            If an empty model is passed (`model.rhs = {}` and `model.algebraic={}`)

        """
        pybamm.logger.info("Start solving {} with {}".format(model.name, self.name))

        # Make sure model isn't empty
        if len(model.rhs) == 0 and len(model.algebraic) == 0:
            raise pybamm.ModelError("Cannot solve empty model")

        # Make sure t_eval is monotonic
        if (np.diff(t_eval) < 0).any():
            raise pybamm.SolverError("t_eval must increase monotonically")

        # Non-dimensionalise t_eval

        # Set up
        timer = pybamm.Timer()

        # Set up external variables and inputs
        external_variables = external_variables or {}
        inputs = inputs or {}
        ext_and_inputs = {**external_variables, **inputs}

        # Raise warning if t_eval looks like it was supposed to be dimensionless
        # already
        if t_eval[-1] < 0.5:
            raise pybamm.SolverError(
                """It looks like t_eval might be dimensionless.
                t_eval should now be provided in seconds"""
            )

        # Set up (if not done already)
        if model not in self.models_set_up:
            self.set_up(model, ext_and_inputs)
            set_up_time = timer.time()
            self.models_set_up.add(model)
        else:
            set_up_time = 0
        # Non-dimensionalise time
        t_eval_dimensionless = t_eval / model.timescale_eval
        # Solve
        # Set inputs and external
        self.set_inputs(model, ext_and_inputs)

        # Calculate discontinuities
        discontinuities = [
            event.expression.evaluate(u=inputs)
            for event in model.discontinuity_events_eval
        ]

        # make sure they are increasing in time
        discontinuities = sorted(discontinuities)

        # remove any identical discontinuities
        discontinuities = [
            v
            for i, v in enumerate(discontinuities)
            if (
                i == len(discontinuities) - 1
                or discontinuities[i] < discontinuities[i + 1]
            )
            and v > 0
        ]

        if len(discontinuities) > 0:
            pybamm.logger.info(
                "Discontinuity events found at t = {}".format(discontinuities)
            )
        else:
            pybamm.logger.info("No discontinuity events found")

        # insert time points around discontinuities in t_eval
        # keep track of sub sections to integrate by storing start and end indices
        start_indices = [0]
        end_indices = []
        eps = sys.float_info.epsilon
        for dtime in discontinuities:
            dindex = np.searchsorted(t_eval_dimensionless, dtime, side="left")
            end_indices.append(dindex + 1)
            start_indices.append(dindex + 1)
            if dtime - eps < t_eval_dimensionless[dindex] < dtime + eps:
                t_eval_dimensionless[dindex] += eps
                t_eval_dimensionless = np.insert(
                    t_eval_dimensionless, dindex, dtime - eps
                )
            else:
                t_eval_dimensionless = np.insert(
                    t_eval_dimensionless, dindex, [dtime - eps, dtime + eps]
                )
        end_indices.append(len(t_eval_dimensionless))

        # integrate separately over each time segment and accumulate into the solution
        # object, restarting the solver at each discontinuity (and recalculating a
        # consistent state afterwards if a dae)
        old_y0 = model.y0
        solution = None
        for start_index, end_index in zip(start_indices, end_indices):
            pybamm.logger.info(
                "Calling solver for {} < t < {}".format(
                    t_eval_dimensionless[start_index] * model.timescale_eval,
                    t_eval_dimensionless[end_index - 1] * model.timescale_eval,
                )
            )
            timer.reset()
            new_solution = self._integrate(
                model, t_eval_dimensionless[start_index:end_index], ext_and_inputs
            )
            new_solution.solve_time = timer.time()
            if solution is None:
                solution = new_solution
            else:
                solution.append(new_solution, start_index=0)

            if solution.termination != "final time":
                break

            if end_index != len(t_eval_dimensionless):
                # setup for next integration subsection
                last_state = solution.y[:, -1]
                if len(model.algebraic) > 0:
                    model.y0 = self.calculate_consistent_state(
                        model,
                        t_eval_dimensionless[end_index],
                        last_state,
                        ext_and_inputs,
                    )
                else:
                    model.y0 = last_state

        # restore old y0
        model.y0 = old_y0

        # Assign times
        solution.set_up_time = set_up_time

        # Add model and inputs to solution
        solution.model = model
        solution.inputs = ext_and_inputs

        # Identify the event that caused termination
        termination = self.get_termination_reason(solution, model.events)

        pybamm.logger.info("Finish solving {} ({})".format(model.name, termination))
        pybamm.logger.info(
            "Set-up time: {}, Solve time: {}, Total time: {}".format(
                timer.format(solution.set_up_time),
                timer.format(solution.solve_time),
                timer.format(solution.total_time),
            )
        )
        return solution

    def step(
        self,
        old_solution,
        model,
        dt,
        npts=2,
        external_variables=None,
        inputs=None,
        save=True,
    ):
        """
        Step the solution of the model forward by a given time increment. The
        first time this method is called it executes the necessary setup by
        calling `self.set_up(model)`.

        Parameters
        ----------
        old_solution : :class:`pybamm.Solution` or None
            The previous solution to be added to. If `None`, a new solution is created.
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate. Must have attributes rhs and
            initial_conditions
        dt : numeric type
            The timestep over which to step the solution
        npts : int, optional
            The number of points at which the solution will be returned during
            the step dt. default is 2 (returns the solution at t0 and t0 + dt).
        external_variables : dict
            A dictionary of external variables and their corresponding
            values at the current time
        inputs : dict, optional
            Any input parameters to pass to the model when solving
        save : bool
            Turn on to store the solution of all previous timesteps

        Raises
        ------
        :class:`pybamm.ModelError`
            If an empty model is passed (`model.rhs = {}` and `model.algebraic={}`)

        """

        if old_solution is not None and not (
            old_solution.termination == "final time"
            or "[experiment]" in old_solution.termination
        ):
            # Return same solution as an event has already been triggered
            # With hack to allow stepping past experiment current / voltage cut-off
            return old_solution

        # Make sure model isn't empty
        if len(model.rhs) == 0 and len(model.algebraic) == 0:
            raise pybamm.ModelError("Cannot step empty model")

        # Set timer
        timer = pybamm.Timer()

        # Set up external variables and inputs
        external_variables = external_variables or {}
        inputs = inputs or {}
        ext_and_inputs = {**external_variables, **inputs}

        # Run set up on first step
        if old_solution is None:
            pybamm.logger.info(
                "Start stepping {} with {}".format(model.name, self.name)
            )
            self.set_up(model, ext_and_inputs)
            t = 0.0
            set_up_time = timer.time()
        else:
            # initialize with old solution
            t = old_solution.t[-1]
            model.y0 = old_solution.y[:, -1]
            set_up_time = 0

        # Non-dimensionalise dt
        dt_dimensionless = dt / model.timescale_eval
        # Step
        t_eval = np.linspace(t, t + dt_dimensionless, npts)
        # Set inputs and external
        self.set_inputs(model, ext_and_inputs)

        pybamm.logger.info("Calling solver")
        timer.reset()
        solution = self._integrate(model, t_eval, ext_and_inputs)

        # Assign times
        solution.set_up_time = set_up_time
        solution.solve_time = timer.time()

        # Add model and inputs to solution
        solution.model = model
        solution.inputs = ext_and_inputs

        # Identify the event that caused termination
        termination = self.get_termination_reason(solution, model.events)

        pybamm.logger.debug("Finish stepping {} ({})".format(model.name, termination))
        if set_up_time:
            pybamm.logger.debug(
                "Set-up time: {}, Step time: {}, Total time: {}".format(
                    timer.format(solution.set_up_time),
                    timer.format(solution.solve_time),
                    timer.format(solution.total_time),
                )
            )
        else:
            pybamm.logger.debug(
                "Step time: {}".format(timer.format(solution.solve_time))
            )
        if save is False or old_solution is None:
            return solution
        else:
            return old_solution + solution

    def get_termination_reason(self, solution, events):
        """
        Identify the cause for termination. In particular, if the solver terminated
        due to an event, (try to) pinpoint which event was responsible.
        Note that the current approach (evaluating all the events and then finding which
        one is smallest at the final timestep) is pretty crude, but is the easiest one
        that works for all the different solvers.

        Parameters
        ----------
        solution : :class:`pybamm.Solution`
            The solution object
        events : dict
            Dictionary of events
        """
        if solution.termination == "final time":
            return "the solver successfully reached the end of the integration interval"
        elif solution.termination == "event":
            # Get final event value
            final_event_values = {}

            for event in events:
                if event.event_type == pybamm.EventType.TERMINATION:
                    final_event_values[event.name] = abs(
                        event.expression.evaluate(
                            solution.t_event,
                            solution.y_event,
                            {k: v[-1] for k, v in solution.inputs.items()},
                        )
                    )
            termination_event = min(final_event_values, key=final_event_values.get)
            # Add the event to the solution object
            solution.termination = "event: {}".format(termination_event)
            return "the termination event '{}' occurred".format(termination_event)


class SolverCallable:
    "A class that will be called by the solver when integrating"

    def __init__(self, function, name, model):
        self._function = function
        if isinstance(function, casadi.Function):
            self.form = "casadi"
            self.inputs = casadi.DM()
        else:
            self.form = "python"
            self.inputs = {}
        self.name = name
        self.model = model

    def set_inputs(self, inputs):
        "Set inputs"
        if self.form == "python":
            self.inputs = inputs
        elif self.form == "casadi":
            self.inputs = casadi.vertcat(*[x for x in inputs.values()])
        self.timescale = self.model.timescale_eval

    def __call__(self, t, y):
        y = y[:, np.newaxis]
        if self.name in ["RHS", "algebraic", "residuals", "event"]:
            pybamm.logger.debug(
                "Evaluating {} for {} at t={}".format(
                    self.name, self.model.name, t * self.timescale
                )
            )
            return self.function(t, y).flatten()
        else:
            return self.function(t, y)

    def function(self, t, y):
        if self.form == "casadi":
            if self.name in ["RHS", "algebraic", "residuals", "event"]:
                return self._function(t, y, self.inputs).full()
            else:
                # keep jacobians sparse
                return self._function(t, y, self.inputs)
        else:
            return self._function(t, y, self.inputs, known_evals={})[0]


class Residuals(SolverCallable):
    "Returns information about residuals at time t and state y"

    def __init__(self, function, name, model):
        super().__init__(function, name, model)
        self.mass_matrix = model.mass_matrix.entries

    def __call__(self, t, y, ydot):
        states_eval = super().__call__(t, y)
        return states_eval - self.mass_matrix @ ydot



#
# Spatial method for zero dimensional meshes
#
import pybamm
import numpy as np


class ZeroDimensionalMethod(pybamm.SpatialMethod):
    """
    A discretisation class for the zero dimensional mesh

    Parameters
    ----------
    mesh : :class: `pybamm.Mesh`
        Contains all the submeshes for discretisation

    **Extends** : :class:`pybamm.SpatialMethod`
    """

    def __init__(self, options=None):
        super().__init__(options)

    def build(self, mesh):
        self._mesh = mesh

    def boundary_value_or_flux(self, symbol, discretised_child, bcs=None):
        """
        In 0D, the boundary value is the identity operator.
        See :meth:`SpatialMethod.boundary_value_or_flux`
        """
        return discretised_child

    def mass_matrix(self, symbol, boundary_conditions):
        """
        Calculates the mass matrix for a spatial method. Since the spatial method is
        zero dimensional, this is simply the number 1.
        """
        return pybamm.Matrix(np.ones((1, 1)))



#
# Geometry class for storing the geometry of the model
#
import pybamm


class Geometry(dict):

    """
    A geometry class to store the details features of the cell geometry.

    Geometry extends the class dictionary and uses the key words: "negative electrode",
    "positive electrode", etc to indicate the subdomain.  Within each subdomain, there
    are "primary", "secondary" or "tabs" dimensions.  "primary" dimensions correspond to
    dimensions on which spatial operators will be applied (e.g. the gradient and
    divergence). In contrast, spatial operators do not act along "secondary" dimensions.
    This allows for multiple independent particles to be included into a model.

    The values assigned to each domain are dictionaries containing the spatial variables
    in that domain, along with expression trees giving their min and maximum extents.
    For example, the following dictionary structure would represent a Geometry with a
    single domain "negative electrode", defined using the variable `x_n` which has a
    range from 0 to the pre-defined parameter `l_n`.

    .. code-block:: python

       {
           "negative electrode": {
               "primary": {x_n: {"min": pybamm.Scalar(0), "max": l_n}}
           }
       }

    A user can create a new Geometry by combining one or more of the pre-defined
    geometries defined with the names given below.

    - "1D macro": macroscopic 1D cell geometry (i.e. electrodes)
    - "3D macro": macroscopic 3D cell geometry
    - "1+1D macro": 1D macroscopic cell geometry with a 1D current collector
    - "1+2D macro": 1D macroscopic cell geometry with a 2D current collector
    - "1D micro": 1D microscopic cell geometry (i.e. particles)
    - "1+1D micro": This is the geometry used in the standard DFN or P2D model
    - "(1+0)+1D micro": 0D macroscopic cell geometry with 1D current collector,
                        along with the microscopic 1D particle geometry.
    - "(2+0)+1D micro": 0D macroscopic cell geometry with 1D current collector,
                        along with the microscopic 1D particle geometry.
    - "(1+1)+1D micro": 1D macroscopic cell geometry, with 1D current collector model,
                        along with the microscopic 1D particle geometry.
    - "(2+1)+1D micro": 1D macroscopic cell geometry, with 2D current collector model,
                        along with the microscopic 1D particle geometry.
    - "2D current collector": macroscopic 2D current collector geometry

    **Extends**: :class:`dict`

    Parameters
    ----------

    geometries: one or more strings or Geometry objects. A string will be assumed to be
                one of the predefined Geometries given above

    custom_geometry : dict containing any extra user defined geometry
    """

    def __init__(self, *geometries, custom_geometry={}):
        for geometry in geometries:
            if geometry == "1D macro":
                geometry = Geometry1DMacro()
            elif geometry == "3D macro":
                geometry = Geometry3DMacro()
            elif geometry == "1+1D macro":
                geometry = Geometryxp1DMacro(cc_dimension=1)
            elif geometry == "2+1D macro":
                geometry = Geometryxp1DMacro(cc_dimension=2)
            elif geometry == "1D micro":
                geometry = Geometry1DMicro()
            elif geometry == "1+1D micro":
                geometry = Geometry1p1DMicro()
            elif geometry == "(1+0)+1D micro":
                geometry = Geometryxp0p1DMicro(cc_dimension=1)
            elif geometry == "(2+0)+1D micro":
                geometry = Geometryxp0p1DMicro(cc_dimension=2)
            elif geometry == "(1+1)+1D micro":
                geometry = Geometryxp1p1DMicro(cc_dimension=1)
            elif geometry == "(2+1)+1D micro":
                geometry = Geometryxp1p1DMicro(cc_dimension=2)
            elif geometry == "2D current collector":
                geometry = Geometry2DCurrentCollector()
            # avoid combining geometries that clash
            if any([k in self.keys() for k in geometry.keys()]):
                raise ValueError("trying to overwrite existing geometry")

            for k, v in geometry.items():
                self.add_domain(k, v)

        # Allow overwriting with a custom geometry
        for k, v in custom_geometry.items():
            self.add_domain(k, v)

    def add_domain(self, name, geometry):
        """
        Add a new domain to the geometry

        Parameters
        ----------

        name: string giving the name of the domain

        geometry: dict of variables in the domain, along with the minimum and maximum
                extents (e.g. {"primary": {x_n: {"min": pybamm.Scalar(0), "max": l_n}}}
        """
        if not isinstance(name, str):
            raise ValueError("name must be a string")

        for k, v in geometry.items():
            if k not in ["primary", "secondary", "tabs"]:
                raise ValueError(
                    'keys of geometry must be either "primary", "secondary" or '
                    '"tabs"'
                )
            if k != "tabs":
                for variable, rnge in v.items():
                    if not isinstance(variable, pybamm.SpatialVariable):
                        raise ValueError(
                            "inner dict of geometry must have pybamm.SpatialVariable "
                            "as keys"
                        )
                    if list(rnge.keys()) != ["position"]:
                        if "min" not in rnge.keys():
                            raise ValueError(
                                "no minimum extents for variable {}".format(variable)
                            )
                        if "max" not in rnge.keys():
                            raise ValueError(
                                "no maximum extents for variable {}".format(variable)
                            )
            else:
                for region, params in v.items():
                    if region not in ["negative", "positive"]:
                        raise ValueError('tabs region must be "negative" or "positive"')
                    for pname in params.keys():
                        if pname not in ["y_centre", "z_centre", "width"]:
                            raise ValueError(
                                'tabs region params must be "y_centre", '
                                '"z_centre" or "width"'
                            )

        self.update({name: geometry})


class Geometry1DMacro(Geometry):
    """
    A geometry class to store the details features of the macroscopic 1D cell geometry.

    **Extends**: :class:`Geometry`

    Parameters
    ----------

    custom_geometry : dict containing any extra user defined geometry
    """

    def __init__(self, custom_geometry={}):
        super().__init__()
        var = pybamm.standard_spatial_vars
        l_n = pybamm.geometric_parameters.l_n
        l_s = pybamm.geometric_parameters.l_s

        self["negative electrode"] = {
            "primary": {var.x_n: {"min": pybamm.Scalar(0), "max": l_n}}
        }
        self["separator"] = {"primary": {var.x_s: {"min": l_n, "max": l_n + l_s}}}
        self["positive electrode"] = {
            "primary": {var.x_p: {"min": l_n + l_s, "max": pybamm.Scalar(1)}}
        }
        self["current collector"] = {"primary": {var.z: {"position": pybamm.Scalar(1)}}}

        # update with custom geometry if non empty
        self.update(custom_geometry)


class Geometry3DMacro(Geometry1DMacro):
    """
    A geometry class to store the details features of the macroscopic 3D cell geometry.

    **Extends**: :class:`Geometry1DMacro`

    Parameters
    ----------

    custom_geometry : dict containing any extra user defined geometry
    """

    def __init__(self, custom_geometry={}):
        super().__init__()

        var = pybamm.standard_spatial_vars

        y_lim = {"min": pybamm.Scalar(0), "max": pybamm.geometric_parameters.l_y}

        z_lim = {"min": pybamm.Scalar(0), "max": pybamm.geometric_parameters.l_z}

        for domain in self.keys():
            self[domain]["primary"][var.y] = y_lim
            self[domain]["primary"][var.z] = z_lim
        self.update(custom_geometry)


class Geometry1DMicro(Geometry):
    """
    A geometry class to store the details features of the microscopic 1D particle
    geometry.

    **Extends**: :class:`Geometry`

    Parameters
    ----------

    custom_geometry : dict containing any extra user defined geometry
    """

    def __init__(self, custom_geometry={}):
        super().__init__()

        var = pybamm.standard_spatial_vars

        self["negative particle"] = {
            "primary": {var.r_n: {"min": pybamm.Scalar(0), "max": pybamm.Scalar(1)}}
        }
        self["positive particle"] = {
            "primary": {var.r_p: {"min": pybamm.Scalar(0), "max": pybamm.Scalar(1)}}
        }
        # update with custom geometry if non empty
        self.update(custom_geometry)


class Geometry1p1DMicro(Geometry):
    """
    A geometry class to store the details features of the 1+1D cell geometry.
    This is the geometry used in the standard DFN or P2D model.

    **Extends**: :class:`Geometry`

    Parameters
    ----------

    custom_geometry : dict containing any extra user defined geometry
    """

    def __init__(self, custom_geometry={}):
        super().__init__()

        var = pybamm.standard_spatial_vars
        l_n = pybamm.geometric_parameters.l_n
        l_s = pybamm.geometric_parameters.l_s

        self["negative particle"] = {
            "primary": {var.r_n: {"min": pybamm.Scalar(0), "max": pybamm.Scalar(1)}},
            "secondary": {var.x_n: {"min": pybamm.Scalar(0), "max": l_n}},
        }
        self["positive particle"] = {
            "primary": {var.r_p: {"min": pybamm.Scalar(0), "max": pybamm.Scalar(1)}},
            "secondary": {var.x_p: {"min": l_n + l_s, "max": pybamm.Scalar(1)}},
        }
        # update with custom geometry if non empty
        self.update(custom_geometry)


class Geometryxp1DMacro(Geometry1DMacro):
    """
    A geometry class to store the details features of x+1D macroscopic cell
    geometry, where x is the dimension of the current collector model.

    **Extends**: :class:`Geometry1DMacro`

    Parameters
    ----------

    cc_dimension : int, optional
        the dimension of the current collector model
    custom_geometry : dict, optional
        dictionary containing any extra user defined geometry
    """

    def __init__(self, cc_dimension=1, custom_geometry={}):
        super().__init__()

        var = pybamm.standard_spatial_vars

        if cc_dimension == 1:
            # Add secondary domains to x-domains
            for geom in self.values():
                geom["secondary"] = {
                    var.z: {
                        "min": pybamm.Scalar(0),
                        "max": pybamm.geometric_parameters.l_z,
                    }
                }
            # Add primary current collector domain
            self["current collector"] = {
                "primary": {
                    var.z: {
                        "min": pybamm.Scalar(0),
                        "max": pybamm.geometric_parameters.l_z,
                    }
                },
                "tabs": {
                    "negative": {
                        "z_centre": pybamm.geometric_parameters.centre_z_tab_n
                    },
                    "positive": {
                        "z_centre": pybamm.geometric_parameters.centre_z_tab_p
                    },
                },
            }
        elif cc_dimension == 2:
            # Add secondary domains to x-domains
            for geom in self.values():
                geom["secondary"] = {
                    var.y: {
                        "min": pybamm.Scalar(0),
                        "max": pybamm.geometric_parameters.l_y,
                    },
                    var.z: {
                        "min": pybamm.Scalar(0),
                        "max": pybamm.geometric_parameters.l_z,
                    },
                }
            # Add primary current collector domain
            self["current collector"] = {
                "primary": {
                    var.y: {
                        "min": pybamm.Scalar(0),
                        "max": pybamm.geometric_parameters.l_y,
                    },
                    var.z: {
                        "min": pybamm.Scalar(0),
                        "max": pybamm.geometric_parameters.l_z,
                    },
                },
                "tabs": {
                    "negative": {
                        "y_centre": pybamm.geometric_parameters.centre_y_tab_n,
                        "z_centre": pybamm.geometric_parameters.centre_z_tab_n,
                        "width": pybamm.geometric_parameters.l_tab_n,
                    },
                    "positive": {
                        "y_centre": pybamm.geometric_parameters.centre_y_tab_p,
                        "z_centre": pybamm.geometric_parameters.centre_z_tab_p,
                        "width": pybamm.geometric_parameters.l_tab_p,
                    },
                },
            }
        else:
            raise pybamm.GeometryError(
                "current collector dimension must be 1 or 2, not {}".format(
                    cc_dimension
                )
            )

        # update with custom geometry if non empty
        self.update(custom_geometry)


class Geometryxp0p1DMicro(Geometry1DMicro):
    """
    A geometry class to store the details features of x+0D macroscopic cell
    geometry, where x is the dimension of the current collector model, along with
    the microscopic 1D particle geometry.

    **Extends**: :class:`Geometry1DMicro`

    Parameters
    ----------

    cc_dimension : int, optional
        the dimension of the current collector model
    custom_geometry : dict, optional
        dictionary containing any extra user defined geometry
    """

    def __init__(self, cc_dimension=1, custom_geometry={}):
        super().__init__()

        var = pybamm.standard_spatial_vars

        # Add secondary domains to x-domains
        if cc_dimension == 1:
            for domain in self.keys():
                self[domain]["secondary"] = {
                    var.z: {
                        "min": pybamm.Scalar(0),
                        "max": pybamm.geometric_parameters.l_z,
                    }
                }
        elif cc_dimension == 2:
            for domain in self.keys():
                self[domain]["secondary"] = {
                    var.y: {
                        "min": pybamm.Scalar(0),
                        "max": pybamm.geometric_parameters.l_y,
                    },
                    var.z: {
                        "min": pybamm.Scalar(0),
                        "max": pybamm.geometric_parameters.l_z,
                    },
                }
        else:
            raise pybamm.GeometryError(
                "current collector dimension must be 1 or 2, not {}".format(
                    cc_dimension
                )
            )

        # update with custom geometry if non empty
        self.update(custom_geometry)


class Geometryxp1p1DMicro(Geometry1DMicro):
    """
    A geometry class to store the details features of x+1D macroscopic cell
    geometry, where x is the dimension of the current collector model, along with
    the microscopic 1D particle geometry.

    **Extends**: :class:`Geometry1DMicro`

    Parameters
    ----------

    cc_dimension : int, optional
        the dimension of the current collector model
    custom_geometry : dict, optional
        dictionary containing any extra user defined geometry
    """

    def __init__(self, cc_dimension=1, custom_geometry={}):
        super().__init__()

        var = pybamm.standard_spatial_vars
        l_n = pybamm.geometric_parameters.l_n
        l_s = pybamm.geometric_parameters.l_s

        # Add secondary domains to x-domains
        if cc_dimension == 1:
            self["negative particle"]["secondary"] = {
                var.x_n: {"min": pybamm.Scalar(0), "max": l_n},
                var.z: {
                    "min": pybamm.Scalar(0),
                    "max": pybamm.geometric_parameters.l_z,
                },
            }
            self["positive particle"]["secondary"] = {
                var.x_p: {"min": l_n + l_s, "max": pybamm.Scalar(1)},
                var.z: {
                    "min": pybamm.Scalar(0),
                    "max": pybamm.geometric_parameters.l_z,
                },
            }
        elif cc_dimension == 2:
            self["negative particle"]["secondary"] = {
                var.x_n: {"min": pybamm.Scalar(0), "max": l_n},
                var.y: {
                    "min": pybamm.Scalar(0),
                    "max": pybamm.geometric_parameters.l_y,
                },
                var.z: {
                    "min": pybamm.Scalar(0),
                    "max": pybamm.geometric_parameters.l_z,
                },
            }
            self["positive particle"]["secondary"] = {
                var.x_p: {"min": l_n + l_s, "max": pybamm.Scalar(1)},
                var.y: {
                    "min": pybamm.Scalar(0),
                    "max": pybamm.geometric_parameters.l_y,
                },
                var.z: {
                    "min": pybamm.Scalar(0),
                    "max": pybamm.geometric_parameters.l_z,
                },
            }
        else:
            raise pybamm.GeometryError(
                "current collector dimension must be 1 or 2, not {}".format(
                    cc_dimension
                )
            )

        # update with custom geometry if non empty
        self.update(custom_geometry)


class Geometry2DCurrentCollector(Geometry):
    """
    A geometry class to store the details features of the macroscopic 2D
    current collector geometry.

    **Extends**: :class:`Geometry`

    Parameters
    ----------

    custom_geometry : dict containing any extra user defined geometry
    """

    def __init__(self, custom_geometry={}):
        super().__init__()
        var = pybamm.standard_spatial_vars

        self["current collector"] = {
            "primary": {
                var.y: {
                    "min": pybamm.Scalar(0),
                    "max": pybamm.geometric_parameters.l_y,
                },
                var.z: {
                    "min": pybamm.Scalar(0),
                    "max": pybamm.geometric_parameters.l_z,
                },
            },
            "tabs": {
                "negative": {
                    "y_centre": pybamm.geometric_parameters.centre_y_tab_n,
                    "z_centre": pybamm.geometric_parameters.centre_z_tab_n,
                    "width": pybamm.geometric_parameters.l_tab_n,
                },
                "positive": {
                    "y_centre": pybamm.geometric_parameters.centre_y_tab_p,
                    "z_centre": pybamm.geometric_parameters.centre_z_tab_p,
                    "width": pybamm.geometric_parameters.l_tab_p,
                },
            },
        }

        # update with custom geometry if non empty
        self.update(custom_geometry)



#
# Solver class using Scipy's adaptive time stepper
#
import pybamm

import scipy.integrate as it
import numpy as np


class ScipySolver(pybamm.BaseSolver):
    """Solve a discretised model, using scipy._integrate.solve_ivp.

    Parameters
    ----------
    method : str, optional
        The method to use in solve_ivp (default is "BDF")
    rtol : float, optional
        The relative tolerance for the solver (default is 1e-6).
    atol : float, optional
        The absolute tolerance for the solver (default is 1e-6).
    """

    def __init__(self, method="BDF", rtol=1e-6, atol=1e-6):
        super().__init__(method, rtol, atol)
        self.ode_solver = True
        self.name = "Scipy solver ({})".format(method)
        pybamm.citations.register("virtanen2020scipy")

    def _integrate(self, model, t_eval, inputs=None):
        """
        Solve a model defined by dydt with initial conditions y0.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate.
        t_eval : :class:`numpy.array`, size (k,)
            The times at which to compute the solution
        inputs : dict, optional
            Any input parameters to pass to the model when solving

        Returns
        -------
        object
            An object containing the times and values of the solution, as well as
            various diagnostic messages.

        """
        extra_options = {"rtol": self.rtol, "atol": self.atol}

        # check for user-supplied Jacobian
        implicit_methods = ["Radau", "BDF", "LSODA"]
        if np.any([self.method in implicit_methods]):
            if model.jacobian_eval:
                extra_options.update({"jac": model.jacobian_eval})

        # make events terminal so that the solver stops when they are reached
        if model.terminate_events_eval:
            for event in model.terminate_events_eval:
                event.terminal = True
            extra_options.update({"events": model.terminate_events_eval})

        sol = it.solve_ivp(
            model.rhs_eval,
            (t_eval[0], t_eval[-1]),
            model.y0,
            t_eval=t_eval,
            method=self.method,
            dense_output=True,
            **extra_options
        )

        if sol.success:
            # Set the reason for termination
            if sol.message == "A termination event occurred.":
                termination = "event"
                t_event = []
                for time in sol.t_events:
                    if time.size > 0:
                        t_event = np.append(t_event, np.max(time))
                t_event = np.array([np.max(t_event)])
                y_event = sol.sol(t_event)
            elif sol.message.startswith("The solver successfully reached the end"):
                termination = "final time"
                t_event = None
                y_event = np.array(None)
            return pybamm.Solution(sol.t, sol.y, t_event, y_event, termination)
        else:
            raise pybamm.SolverError(sol.message)



#
# Bulter volmer class
#

import pybamm
from ..base_interface import BaseInterface


class BaseInverseKinetics(BaseInterface):
    """
    A base submodel that implements the inverted form of the Butler-Volmer relation to
    solve for the reaction overpotential.

    Parameters
    ----------
    param
        Model parameters
    domain : iter of str, optional
        The domain(s) in which to compute the interfacial current. Default is None,
        in which case j.domain is used.

    **Extends:** :class:`pybamm.interface.kinetics.ButlerVolmer`

    """

    def __init__(self, param, domain):
        super().__init__(param, domain)

    def get_coupled_variables(self, variables):
        ocp, dUdT = self._get_open_circuit_potential(variables)

        j0 = self._get_exchange_current_density(variables)
        j_tot_av = self._get_average_total_interfacial_current_density(variables)
        # Broadcast to match j0's domain
        if j0.domain in [[], ["current collector"]]:
            j = j_tot_av
        else:
            j = pybamm.PrimaryBroadcast(j_tot_av, [self.domain.lower() + " electrode"])

        if self.domain == "Negative":
            ne = self.param.ne_n
        elif self.domain == "Positive":
            ne = self.param.ne_p
        # Note: T must have the same domain as j0 and eta_r
        if j0.domain in ["current collector", ["current collector"]]:
            T = variables["X-averaged cell temperature"]
        else:
            T = variables[self.domain + " electrode temperature"]

        eta_r = self._get_overpotential(j, j0, ne, T)
        delta_phi = eta_r + ocp

        variables.update(self._get_standard_interfacial_current_variables(j))
        variables.update(
            self._get_standard_total_interfacial_current_variables(j_tot_av)
        )
        variables.update(self._get_standard_exchange_current_variables(j0))
        variables.update(self._get_standard_overpotential_variables(eta_r))
        variables.update(
            self._get_standard_surface_potential_difference_variables(delta_phi)
        )
        variables.update(self._get_standard_ocp_variables(ocp, dUdT))

        if (
            "Negative electrode" + self.reaction_name + " interfacial current density"
            in variables
            and "Positive electrode"
            + self.reaction_name
            + " interfacial current density"
            in variables
        ):
            variables.update(
                self._get_standard_whole_cell_interfacial_current_variables(variables)
            )
            variables.update(
                self._get_standard_whole_cell_exchange_current_variables(variables)
            )

        return variables

    def _get_overpotential(self, j, j0, ne, T):
        raise NotImplementedError



from pybamm import exp


def nmc_LGM50_diffusivity_Chen2020(sto, T, T_inf, E_D_s, R_g):
    """
       NMC diffusivity as a function of stoichiometry, in this case the
       diffusivity is taken to be a constant. The value is taken from [1].
       References
       ----------
       .. [1] Chang-Hui Chen, Ferran Brosa Planella, Kieran O’Regan, Dominika Gastol,
       W. Dhammika Widanage, and Emma Kendrick. "Development of Experimental Techniques
       for Parameterization of Multi-scale Lithium-ion Battery Models." Submitted for
       publication (2020).
       Parameters
       ----------
       sto: :class: `numpy.Array`
         Electrode stochiometry
       T: :class: `numpy.Array`
          Dimensional temperature
       T_inf: double
          Reference temperature
       E_D_s: double
          Solid diffusion activation energy
       R_g: double
          The ideal gas constant
       Returns
       -------
       : double
          Solid diffusivity
    """

    D_ref = 4e-15
    arrhenius = exp(E_D_s / R_g * (1 / T_inf - 1 / T))

    return D_ref * arrhenius



from .base_convection import BaseModel
from .no_convection import NoConvection
from .leading_convection import LeadingOrder
from .full_convection import Full
from .composite_convection import Composite



#
# Class for one-dimensional current collectors in which the potential is held
# fixed and the current is determined from the I-V relationship used in the SPM(e)
#
import pybamm
from .base_current_collector import BaseModel


class BaseSetPotentialSingleParticle(BaseModel):
    """A submodel for current collectors which *doesn't* update the potentials
    during solve. This class uses the current-voltage relationship from the
    SPM(e) (see [1]_) to calculate the current.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel

    References
    ----------
    .. [1] SG Marquis, V Sulzer, R Timms, CP Please and SJ Chapman. “An asymptotic
           derivation of a single particle model with electrolyte”. In: arXiv preprint
           arXiv:1905.12553 (2019).


    **Extends:** :class:`pybamm.current_collector.BaseModel`
    """

    def __init__(self, param):
        super().__init__(param)

    def get_fundamental_variables(self):

        phi_s_cn = pybamm.standard_variables.phi_s_cn

        variables = self._get_standard_negative_potential_variables(phi_s_cn)

        # TO DO: grad not implemented for 2D yet
        i_cc = pybamm.Scalar(0)
        i_boundary_cc = pybamm.standard_variables.i_boundary_cc

        variables.update(self._get_standard_current_variables(i_cc, i_boundary_cc))
        # Hack to get the leading-order current collector current density
        # Note that this should be different from the actual (composite) current
        # collector current density for 2+1D models, but not sure how to implement this
        # using current structure of lithium-ion models
        variables["Leading-order current collector current density"] = variables[
            "Current collector current density"
        ]

        return variables

    def set_rhs(self, variables):
        phi_s_cn = variables["Negative current collector potential"]

        # Dummy equations so that PyBaMM doesn't change the potentials during solve
        # i.e. d_phi/d_t = 0. Potentials are set externally between steps.
        self.rhs = {phi_s_cn: pybamm.Scalar(0)}

    def set_algebraic(self, variables):
        ocp_p_av = variables["X-averaged positive electrode open circuit potential"]
        ocp_n_av = variables["X-averaged negative electrode open circuit potential"]
        eta_r_n_av = variables["X-averaged negative electrode reaction overpotential"]
        eta_r_p_av = variables["X-averaged positive electrode reaction overpotential"]
        eta_e_av = variables["X-averaged electrolyte overpotential"]
        delta_phi_s_n_av = variables["X-averaged negative electrode ohmic losses"]
        delta_phi_s_p_av = variables["X-averaged positive electrode ohmic losses"]

        i_boundary_cc = variables["Current collector current density"]
        v_boundary_cc = variables["Local voltage"]
        # The voltage-current expression from the SPM(e)
        local_voltage_expression = (
            ocp_p_av
            - ocp_n_av
            + eta_r_p_av
            - eta_r_n_av
            + eta_e_av
            + delta_phi_s_p_av
            - delta_phi_s_n_av
        )
        self.algebraic = {i_boundary_cc: v_boundary_cc - local_voltage_expression}

    def set_initial_conditions(self, variables):

        applied_current = variables["Total current density"]
        cc_area = self._get_effective_current_collector_area()
        phi_s_cn = variables["Negative current collector potential"]
        i_boundary_cc = variables["Current collector current density"]

        self.initial_conditions = {
            phi_s_cn: pybamm.Scalar(0),
            i_boundary_cc: applied_current / cc_area,
        }


class SetPotentialSingleParticle1plus1D(BaseSetPotentialSingleParticle):
    "Class for 1+1D set potential model"

    def __init__(self, param):
        super().__init__(param)

    def _get_effective_current_collector_area(self):
        "In the 1+1D models the current collector effectively has surface area l_z"
        return self.param.l_z


class SetPotentialSingleParticle2plus1D(BaseSetPotentialSingleParticle):
    "Class for 1+1D set potential model"

    def __init__(self, param):
        super().__init__(param)

    def _get_effective_current_collector_area(self):
        "Return the area of the current collector"
        return self.param.l_y * self.param.l_z



#
# One-dimensional submeshes
#
import pybamm
from .meshes import SubMesh

import numpy as np


class SubMesh1D(SubMesh):
    """
    1D submesh class.
    Contains the position of the nodes, the number of mesh points, and
    (optionally) information about the tab locations.

    Parameters
    ----------
    edges : array_like
        An array containing the points corresponding to the edges of the submesh
    coord_sys : string
        The coordinate system of the submesh
    tabs : dict, optional
        A dictionary that contains information about the size and location of
        the tabs

    **Extends:"": :class:`pybamm.SubMesh`
    """

    def __init__(self, edges, coord_sys, tabs=None):
        self.edges = edges
        self.nodes = (self.edges[1:] + self.edges[:-1]) / 2
        self.d_edges = np.diff(self.edges)
        self.d_nodes = np.diff(self.nodes)
        self.npts = self.nodes.size
        self.coord_sys = coord_sys

        # Add tab locations in terms of "left" and "right"
        if tabs:
            self.tabs = {}
            l_z = self.edges[-1]

            def near(x, point, tol=3e-16):
                return abs(x - point) < tol

            for tab in ["negative", "positive"]:
                tab_location = tabs[tab]["z_centre"]
                if near(tab_location, 0):
                    self.tabs[tab + " tab"] = "left"
                elif near(tab_location, l_z):
                    self.tabs[tab + " tab"] = "right"
                else:
                    raise pybamm.GeometryError(
                        """{} tab located at {}, but must be at either 0 or {}
                        (in dimensionless coordinates).""".format(
                            tab, tab_location, l_z
                        )
                    )


class Uniform1DSubMesh(SubMesh1D):
    """
    A class to generate a uniform submesh on a 1D domain

    Parameters
    ----------
    lims : dict
        A dictionary that contains the limits of the spatial variables
    npts : dict
        A dictionary that contains the number of points to be used on each
        spatial variable. Note: the number of nodes (located at the cell centres)
        is npts, and the number of edges is npts+1.
    tabs : dict, optional
        A dictionary that contains information about the size and location of
        the tabs

    **Extends:"": :class:`pybamm.SubMesh1D`
    """

    def __init__(self, lims, npts, tabs=None):

        # check that only one variable passed in
        if len(lims) != 1:
            raise pybamm.GeometryError("lims should only contain a single variable")

        spatial_var = list(lims.keys())[0]
        spatial_lims = lims[spatial_var]
        npts = npts[spatial_var.id]

        edges = np.linspace(spatial_lims["min"], spatial_lims["max"], npts + 1)

        coord_sys = spatial_var.coord_sys

        super().__init__(edges, coord_sys=coord_sys, tabs=tabs)


class Exponential1DSubMesh(SubMesh1D):
    """
    A class to generate a submesh on a 1D domain in which the points are clustered
    close to one or both of boundaries using an exponential formula on the interval
    [a,b].

    If side is "left", the gridpoints are given by

    .. math::
        x_{k} = (b-a) + \\frac{\\exp{\\alpha k / N} - 1}{\\exp{\\alpha} - 1}} + a,

    for k = 1, ..., N, where N is the number of nodes.

    Is side is "right", the gridpoints are given by

    .. math::
        x_{k} = (b-a) + \\frac{\\exp{-\\alpha k / N} - 1}{\\exp{-\\alpha} - 1}} + a,

    for k = 1, ..., N.

    If side is "symmetric", the first half of the interval is meshed using the
    gridpoints

    .. math::
        x_{k} = (b/2-a) + \\frac{\\exp{\\alpha k / N} - 1}{\\exp{\\alpha} - 1}} + a,

    for k = 1, ..., N. The grid spacing is then reflected to contruct the grid
    on the full interval [a,b].

    In the above, alpha is a stretching factor. As the number of gridpoints tends
    to infinity, the ratio of the largest and smallest grid cells tends to exp(alpha).

    Parameters
    ----------
    lims : dict
        A dictionary that contains the limits of the spatial variables
    npts : dict
        A dictionary that contains the number of points to be used on each
        spatial variable. Note: the number of nodes (located at the cell centres)
        is npts, and the number of edges is npts+1.
    tabs : dict
        A dictionary that contains information about the size and location of
        the tabs
    side : str, optional
        Whether the points are clustered near to the left or right boundary,
        or both boundaries. Can be "left", "right" or "symmetric". Default is
        "symmetric"
    stretch : float, optional
        The factor (alpha) which appears in the exponential. If side is "symmetric"
        then the default stretch is 1.15. If side is "left" or "right" then the
        default stretch is 2.3.

    **Extends:"": :class:`pybamm.SubMesh1D`
    """

    def __init__(self, lims, npts, tabs, side="symmetric", stretch=None):

        # check that only one variable passed in
        if len(lims) != 1:
            raise pybamm.GeometryError("lims should only contain a single variable")

        spatial_var = list(lims.keys())[0]
        spatial_lims = lims[spatial_var]
        a = spatial_lims["min"]
        b = spatial_lims["max"]
        npts = npts[spatial_var.id]
        coord_sys = spatial_var.coord_sys

        # Set stretch if not provided
        if not stretch:
            if side == "symmetric":
                stretch = 1.15
            elif side in ["left", "right"]:
                stretch = 2.3

        # Create edges accoriding to "side"
        if side == "left":
            ii = np.array(range(0, npts + 1))
            edges = (b - a) * (np.exp(stretch * ii / npts) - 1) / (
                np.exp(stretch) - 1
            ) + a

        elif side == "right":
            ii = np.array(range(0, npts + 1))
            edges = (b - a) * (np.exp(-stretch * ii / npts) - 1) / (
                np.exp(-stretch) - 1
            ) + a

        elif side == "symmetric":
            # Mesh half-interval [a, b/2]
            if npts % 2 == 0:
                ii = np.array(range(0, int((npts) / 2)))
            else:
                ii = np.array(range(0, int((npts + 1) / 2)))
            x_exp_left = (b / 2 - a) * (np.exp(stretch * ii / npts) - 1) / (
                np.exp(stretch) - 1
            ) + a

            # Refelct mesh
            x_exp_right = b * np.ones_like(x_exp_left) - (x_exp_left[::-1] - a)

            # Combine left and right halves of the mesh, adding a node at the
            # centre if npts is even (odd number of edges)
            if npts % 2 == 0:
                edges = np.concatenate((x_exp_left, [(a + b) / 2], x_exp_right))
            else:
                edges = np.concatenate((x_exp_left, x_exp_right))

        super().__init__(edges, coord_sys=coord_sys, tabs=tabs)


class Chebyshev1DSubMesh(SubMesh1D):
    """
    A class to generate a submesh on a 1D domain using Chebyshev nodes on the
    interval (a, b), given by

    .. math::
        x_{k} = \\frac{1}{2}(a+b) + \\frac{1}{2}(b-a) \\cos(\\frac{2k-1}{2N}\\pi),

    for k = 1, ..., N, where N is the number of nodes. Note: this mesh then
    appends the boundary edges, so that the mesh edges are given by

    .. math::
        a < x_{1} < ... < x_{N} < b.

    Parameters
    ----------
    lims : dict
        A dictionary that contains the limits of the spatial variables
    npts : dict
        A dictionary that contains the number of points to be used on each
        spatial variable. Note: the number of nodes (located at the cell centres)
        is npts, and the number of edges is npts+1.
    tabs : dict, optional
        A dictionary that contains information about the size and location of
        the tabs

    **Extends:"": :class:`pybamm.SubMesh1D`
    """

    def __init__(self, lims, npts, tabs=None):

        # check that only one variable passed in
        if len(lims) != 1:
            raise pybamm.GeometryError("lims should only contain a single variable")

        spatial_var = list(lims.keys())[0]
        spatial_lims = lims[spatial_var]
        npts = npts[spatial_var.id]

        # Create N Chebyshev nodes in the interval (a,b)
        N = npts - 1
        ii = np.array(range(1, N + 1))
        a = spatial_lims["min"]
        b = spatial_lims["max"]
        x_cheb = (a + b) / 2 + (b - a) / 2 * np.cos((2 * ii - 1) * np.pi / 2 / N)

        # Append the boundary nodes. Note: we need to flip the order the Chebyshev
        # nodes as they are created in descending order.
        edges = np.concatenate(([a], np.flip(x_cheb), [b]))
        coord_sys = spatial_var.coord_sys

        super().__init__(edges, coord_sys=coord_sys, tabs=tabs)


class UserSupplied1DSubMesh(SubMesh1D):
    """
    A class to generate a submesh on a 1D domain from a user supplied array of
    edges.

    Parameters
    ----------
    lims : dict
        A dictionary that contains the limits of the spatial variables
    npts : dict
        A dictionary that contains the number of points to be used on each
        spatial variable. Note: the number of nodes (located at the cell centres)
        is npts, and the number of edges is npts+1.
    tabs : dict
        A dictionary that contains information about the size and location of
        the tabs
    edges : array_like
        The array of points which correspond to the edges of the mesh.

    **Extends:"": :class:`pybamm.SubMesh1D`
    """

    def __init__(self, lims, npts, tabs, edges=None):

        # raise error if no edges passed
        if edges is None:
            raise pybamm.GeometryError("User mesh requires parameter 'edges'")

        # check that only one variable passed in
        if len(lims) != 1:
            raise pybamm.GeometryError("lims should only contain a single variable")

        spatial_var = list(lims.keys())[0]
        spatial_lims = lims[spatial_var]
        npts = npts[spatial_var.id]

        # check that npts + 1 equals number of user-supplied edges
        if (npts + 1) != len(edges):
            raise pybamm.GeometryError(
                """User-suppled edges has should have length (npts + 1) but has length {}.
                 Number of points (npts) for domain {} is {}.""".format(
                    len(edges), spatial_var.domain, npts
                )
            )

        # check end points of edges agree with spatial_lims
        if edges[0] != spatial_lims["min"]:
            raise pybamm.GeometryError(
                """First entry of edges is {}, but should be equal to {}
                 for domain {}.""".format(
                    edges[0], spatial_lims["min"], spatial_var.domain
                )
            )
        if edges[-1] != spatial_lims["max"]:
            raise pybamm.GeometryError(
                """Last entry of edges is {}, but should be equal to {}
                for domain {}.""".format(
                    edges[-1], spatial_lims["max"], spatial_var.domain
                )
            )

        coord_sys = spatial_var.coord_sys

        super().__init__(edges, coord_sys=coord_sys, tabs=tabs)
